"""
Tools for plotting and controlling plot output using context managers

.. autosummary::
   :nosignatures:

   add_scaled_colorbar
   disable_interactive
   plot_on_axes
   plot_on_figure
   PlotReference
   BasicPlottingContext
   JupyterPlottingContext
   get_plotting_context
   napari_add_layers

.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
"""

from __future__ import annotations

import contextlib
import functools
import logging
import sys
import warnings
from typing import TYPE_CHECKING, Any, Dict, Generator, Literal, Optional, Type

from ..tools.docstrings import replace_in_docstring

if TYPE_CHECKING:
    import matplotlib.cm  # @UnusedImport
    import napari  # @UnusedImport

    from ..grids.base import GridBase  # @UnusedImport


def add_scaled_colorbar(
    axes_image: matplotlib.cm.ScalarMappable,
    ax=None,
    aspect: float = 20,
    pad_fraction: float = 0.5,
    label: str = "",
    **kwargs,
):
    """add a vertical color bar to an image plot

    The height of the colorbar is now adjusted to the plot, so that the width
    determined by `aspect` is now given relative to the height. Moreover, the
    gap between the colorbar and the plot is now given in units of the fraction
    of the width by `pad_fraction`.

    Inspired by https://stackoverflow.com/a/33505522/932593

    Args:
        axes_image (:class:`matplotlib.cm.ScalarMappable`):
            Mappable object, e.g., returned from :meth:`matplotlib.pyplot.imshow`
        ax (:class:`matplotlib.axes.Axes`):
            The current figure axes from which space is taken for the colorbar. If
            omitted, the axes in which the `axes_image` is shown is taken.
        aspect (float):
            The target aspect ratio of the colorbar
        pad_fraction (float):
            Width of the gap between colorbar and image
        label (str):
            Set a label for the colorbar
        **kwargs:
            Additional parameters are passed to colorbar call

    Returns:
        :class:`~matplotlib.colorbar.Colorbar`: The resulting Colorbar object
    """
    import matplotlib.pyplot as plt
    from mpl_toolkits import axes_grid1

    class _AxesXY(axes_grid1.axes_size._Base):
        """
        Scaled size whose relative part corresponds to the maximum of the data width and
        data height of the *axes* multiplied by the *aspect*.
        """

        def __init__(self, axes, aspect=1.0):
            self._axes = axes
            self._aspect = aspect

        def get_size(self, renderer):
            l1, l2 = self._axes.get_xlim()
            rel_size_x = abs(l2 - l1) * self._aspect

            l1, l2 = self._axes.get_ylim()
            rel_size_y = abs(l2 - l1) * self._aspect

            abs_size = 0.0
            rel_size = max(rel_size_x, rel_size_y)
            return rel_size, abs_size

    if ax is None:
        ax = axes_image.axes

    # make space for the colorbar and generate its axes
    divider = axes_grid1.make_axes_locatable(ax)
    width = _AxesXY(ax, aspect=1.0 / aspect)
    pad = axes_grid1.axes_size.Fraction(pad_fraction, width)
    cax = divider.append_axes("right", size=width, pad=pad)

    # create the colorbar
    cbar = ax.figure.colorbar(axes_image, cax=cax, **kwargs)

    # disable the offset that matplotlib sometimes shows
    try:
        cax.get_xaxis().get_major_formatter().set_useOffset(False)
    except AttributeError:
        pass  # can happen for logarithmically formatted axes
    try:
        cax.get_yaxis().get_major_formatter().set_useOffset(False)
    except AttributeError:
        pass  # can happen for logarithmically formatted axes

    if label:
        cbar.set_label(label)

    plt.sca(ax)  # ensure that the colorbar is not set as the current axes

    return cbar


class nested_plotting_check:
    """context manager that checks whether it is the root plotting call

    Example:
        The context manager can be used in plotting calls to check for nested
        plotting calls::

            with nested_plotting_check() as is_outermost_plot_call:
                make_plot(...)  # could potentially call other plotting methods
                if is_outermost_plot_call:
                    plt.show()

    """

    _is_plotting = False  # class variable keeping track of nesting

    def __init__(self) -> None:
        self.is_nested = None  # determines whether the this context is nested

    def __enter__(self):
        self.is_nested = self.__class__._is_plotting
        self.__class__._is_plotting = True
        return not self.is_nested

    def __exit__(self, *exc):
        if not self.is_nested:
            self.__class__._is_plotting = False


@contextlib.contextmanager
def disable_interactive():
    """context manager disabling the interactive mode of matplotlib

    This context manager restores the previous state after it is done. Details
    of the interactive mode are described in :func:`matplotlib.interactive`.
    """
    import matplotlib.pyplot as plt

    if plt.isinteractive():
        # interactive mode is enabled => disable it temporarily
        plt.interactive(False)
        yield
        plt.interactive(True)

    else:
        # interactive mode is already disabled => do nothing
        yield


class PlotReference:
    """contains all information to update a plot element"""

    __slots__ = ["ax", "element", "parameters"]

    def __init__(self, ax, element: Any, parameters: Optional[Dict[str, Any]] = None):
        """
        Args:
            ax (:class:`matplotlib.axes.Axes`): The axes of the element
            element (:class:`matplotlib.artist.Artist`): The actual element
            parameters (dict): Parameters to recreate the plot element
        """
        self.ax = ax
        self.element = element
        self.parameters = {} if parameters is None else parameters


PlotActionType = Literal["auto", "close", "none", "sca", "show"]


def plot_on_axes(wrapped=None, update_method=None):
    """decorator for a plot method or function that uses a single axes

    This decorator adds typical options for creating plots that fill a single axes.
    These options are available via keyword arguments. To avoid redundancy in describing
    these options in the docstring, the placeholder `{PLOT_ARGS}` can be added to the
    docstring of the wrapped function or method and will be replaced by the appropriate
    text. Note that the decorator can be used on both functions and methods.

    Example:
        The following example illustrates how this decorator can be used to
        implement plotting for a given class. In particular, supplying the
        `update_method` will allow efficient dynamical plotting::

            class State:
                def __init__(self) -> None:
                    self.data = np.arange(8)

                def _update_plot(self, reference):
                    reference.element.set_ydata(self.data)

                @plot_on_axes(update_method='_update_plot')
                def plot(self, ax):
                    line, = ax.plot(np.arange(8), self.data)
                    return PlotReference(ax, line)


            @plot_on_axes
            def make_plot(ax):
                ax.plot(...)

        When `update_method` is absent, the method can still be used for plotting, but
        dynamic updating, e.g., by :class:`pde.trackers.PlotTracker`, is not possible.

    Args:
        wrapped (callable):
            Function to be wrapped
        update_method (callable or str):
            Method to call to update the plot. The argument of the new
            method will be the result of the initial call of the wrapped
            method.
    """
    if wrapped is None:
        # handle the case where decorator was called without brackets
        return functools.partial(plot_on_axes, update_method=update_method)

    def wrapper(
        *args,
        title: Optional[str] = None,
        filename: Optional[str] = None,
        action: PlotActionType = "auto",
        ax_style: Optional[Dict[str, Any]] = None,
        fig_style: Optional[Dict[str, Any]] = None,
        ax=None,
        **kwargs,
    ):
        """
        title (str):
            Title of the plot. If omitted, the title might be chosen automatically.
        filename (str, optional):
            If given, the plot is written to the specified file.
        action (str):
            Decides what to do with the final figure. If the argument is set to `show`,
            :func:`matplotlib.pyplot.show` will be called to show the plot. If the value
            is `none`, the figure will be created, but not necessarily shown. The value
            `close` closes the figure, after saving it to a file when `filename` is
            given. The default value `auto` implies that the plot is shown if it is not
            a nested plot call.
        ax_style (dict):
            Dictionary with properties that will be changed on the axis after the plot
            has been drawn by calling :meth:`matplotlib.pyplot.setp`. A special item i
            this dictionary is `use_offset`, which is flag that can be used to control
            whether offset are shown along the axes of the plot.
        fig_style (dict):
            Dictionary with properties that will be changed on the figure after the plot
            has been drawn by calling :meth:`matplotlib.pyplot.setp`. For instance,
            using fig_style={'dpi': 200} increases the resolution of the figure.
        ax (:class:`matplotlib.axes.Axes`):
            Figure axes to be used for plotting. The special value "create" creates a
            new figure, while "reuse" attempts to reuse an existing figure, which is
            the default.
        """
        # Note on docstring: This docstring replaces the token {PLOT_ARGS} in
        # the wrapped function
        import matplotlib as mpl
        import matplotlib.pyplot as plt

        if ax_style is None:
            ax_style = {}

        # Show figure by calling `plt.show` automatically only if
        #    - action == 'auto'
        #    - axis is not given, i.e., ax == None
        #    - backend is not `inline` (which would show figure anyway)
        # This safeguard is necessary to allow specifying subplot axes explicitly
        # through the `ax` argument.
        auto_show_figure = ax is None and "backend_inline" not in mpl.get_backend()

        # some logic to check for nested plotting calls:
        with nested_plotting_check() as is_outermost_plot_call:
            # disable interactive plotting temporarily
            with disable_interactive():
                if ax is None:
                    # create new figure
                    backend = mpl.get_backend()
                    if "backend_inline" in backend or "nbAgg" == backend:
                        plt.close("all")  # close left over figures
                        auto_show_figure = True  # show this figure if action == 'auto'
                    fig, ax = plt.subplots()

                elif ax == "reuse":
                    # try to reuse an existing figure (or create a new one)
                    ax = plt.gca()
                    fig = ax.get_figure()

                else:
                    # assume an axes was given
                    fig = ax.get_figure()

                # call the actual plotting function
                reference = wrapped(*args, ax=ax, **kwargs)

                # finishing touches...
                if title is not None:
                    ax.set_title(title)

                use_offset = ax_style.pop("use_offset", False)
                if use_offset is not None:
                    ax.get_xaxis().get_major_formatter().set_useOffset(use_offset)
                    ax.get_yaxis().get_major_formatter().set_useOffset(use_offset)
                if ax_style:
                    plt.setp(ax, **ax_style)
                if fig_style:
                    plt.setp(fig, **fig_style)
                if filename:
                    fig.savefig(filename)

            # decide what to do with the final plot
            if action == "auto":
                if is_outermost_plot_call and auto_show_figure:
                    # only call show on the outermost plot call and only in the
                    # circumstances determined above
                    action = "show"
                else:
                    action = "sca"

            if action == "sca":
                # set the axes as the current axes, so subsequent plot calls modify it
                plt.sca(ax)

            elif action == "show":
                # show the entire figure
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore")
                    plt.show()

            elif action == "close":
                # close the figure, e.g., because it has already been used
                plt.close(fig)

            elif action != "none":
                # do nothing if "none", otherwise raise error
                raise ValueError(f"Unknown action `{action}`")

        return reference

    # adjusting the signature of the wrapped function to include wrapper args
    import inspect

    sig_wrapped = inspect.signature(wrapped)
    parameters = tuple(
        arg
        for name, arg in sig_wrapped.parameters.items()
        if name != "kwargs" and name != "ax"
    )

    sig_wrapper = inspect.signature(wrapper)
    parameters += tuple(sig_wrapper.parameters.values())
    wrapper.__signature__ = sig_wrapped.replace(parameters=parameters)

    # adjusting additional properties of the function to match the wrapped one
    wrapper.__name__ = wrapped.__name__
    wrapper.__module__ = wrapped.__module__
    wrapper.__dict__.update(wrapped.__dict__)

    if wrapped.__doc__:
        replace_in_docstring(
            wrapper, "{PLOT_ARGS}", wrapper.__doc__, docstring=wrapped.__doc__
        )

    wrapper.mpl_class = "axes"
    wrapper.update_method = update_method

    return wrapper


def plot_on_figure(wrapped=None, update_method=None):
    """decorator for a plot method or function that fills an entire figure

    This decorator adds typical options for creating plots that fill an entire figure.
    This decorator adds typical options for creating plots that fill a single axes.
    These options are available via keyword arguments. To avoid redundancy in describing
    these options in the docstring, the placeholder `{PLOT_ARGS}` can be added to the
    docstring of the wrapped function or method and will be replaced by the appropriate
    text. Note that the decorator can be used on both functions and methods.

    Example:
        The following example illustrates how this decorator can be used to
        implement plotting for a given class. In particular, supplying the
        `update_method` will allow efficient dynamical plotting::

            class State:
                def __init__(self) -> None:
                    self.data = np.random.random((2, 8))

                def _update_plot(self, reference):
                    ref1, ref2 = reference
                    ref1.element.set_ydata(self.data[0])
                    ref2.element.set_ydata(self.data[1])

                @plot_on_figure(update_method='_update_plot')
                def plot(self, fig):
                    ax1, ax2 = fig.subplots(1, 2)
                    l1, = ax1.plot(np.arange(8), self.data[0])
                    l2, = ax2.plot(np.arange(8), self.data[1])
                    return [PlotReference(ax1, l1), PlotReference(ax2, l2)]


            @plot_on_figure
            def make_plot(fig):
                ...


        When `update_method` is not supplied, the method can still be used for
        plotting, but dynamic updating, e.g., by
        :class:`pde.trackers.PlotTracker`, is not possible.


    Args:
        wrapped (callable):
            Function to be wrapped
        update_method (callable or str):
            Method to call to update the plot. The argument of the new
            method will be the result of the initial call of the wrapped
            method.
    """
    if wrapped is None:
        # handle the case where decorator was called without brackets
        return functools.partial(plot_on_figure, update_method=update_method)

    def wrapper(
        *args,
        title: Optional[str] = None,
        constrained_layout: bool = True,
        filename: Optional[str] = None,
        action: PlotActionType = "auto",
        fig_style: Optional[Dict[str, Any]] = None,
        fig=None,
        **kwargs,
    ):
        """
        title (str):
            Title of the plot. If omitted, the title might be chosen automatically.
            This is shown above all panels.
        constrained_layout (bool):
            Whether to use `constrained_layout` in :func:`matplotlib.pyplot.figure` call
            to create a figure. This affects the layout of all plot elements. Generally,
            spacing might be better with this flag enabled, but it can also lead to
            problems when plotting multiple plots successively, e.g., when creating a
            movie.
        filename (str, optional):
            If given, the figure is written to the specified file.
        action (str):
            Decides what to do with the final figure. If the argument is set to `show`,
            :func:`matplotlib.pyplot.show` will be called to show the plot. If the value
            is `none`, the figure will be created, but not necessarily shown. The value
            `close` closes the figure, after saving it to a file when `filename` is
            given. The default value `auto` implies that the plot is shown if it is not
            a nested plot call.
        fig_style (dict):
            Dictionary with properties that will be changed on the figure after the plot
            has been drawn by calling :meth:`matplotlib.pyplot.setp`. For instance,
            using fig_style={'dpi': 200} increases the resolution of the figure.
        fig (:class:`matplotlib.figures.Figure`):
            Figure that is used for plotting. If omitted, a new figure is created.
        """
        # Note on docstring: This docstring replaces the token {PLOT_ARGS} in
        # the wrapped function
        import matplotlib as mpl
        import matplotlib.pyplot as plt

        # some logic to check for nested plotting calls:
        with nested_plotting_check() as is_outermost_plot_call:
            # disable interactive plotting temporarily
            with disable_interactive():
                if fig is None:
                    # create new figure
                    backend = mpl.get_backend()
                    if "backend_inline" in backend or "nbAgg" == backend:
                        plt.close("all")  # close left over figures
                    fig = plt.figure(constrained_layout=constrained_layout)

                # call the actual plotting function
                reference = wrapped(*args, fig=fig, **kwargs)

                # finishing touches...
                if title is not None:
                    fig.suptitle(title)
                if fig_style:
                    plt.setp(fig, **fig_style)
                if filename:
                    fig.savefig(filename)

            # decide what to do with the final plot
            if action == "auto":
                if is_outermost_plot_call:
                    # only call show on the outermost plot call
                    action = "show"
                else:
                    action = "none"

            if action == "show":
                # show the entire figure
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore")
                    plt.show()

            elif action == "close":
                # close the figure, e.g., because it has already been used
                plt.close(fig)

            elif action != "none":
                # do nothing if "none", otherwise raise error
                raise ValueError(f"Unknown action `{action}`")

        return reference

    # adjusting the signature of the wrapped function to include wrapper args
    import inspect

    sig_wrapped = inspect.signature(wrapped)
    parameters = tuple(
        arg
        for name, arg in sig_wrapped.parameters.items()
        if name != "kwargs" and name != "fig"
    )

    sig_wrapper = inspect.signature(wrapper)
    parameters += tuple(sig_wrapper.parameters.values())
    wrapper.__signature__ = sig_wrapped.replace(parameters=parameters)

    # adjusting additional properties of the function to match the wrapped one
    wrapper.__name__ = wrapped.__name__
    wrapper.__module__ = wrapped.__module__
    wrapper.__dict__.update(wrapped.__dict__)

    if wrapped.__doc__:
        replace_in_docstring(
            wrapper, "{PLOT_ARGS}", wrapper.__doc__, docstring=wrapped.__doc__
        )

    wrapper.mpl_class = "figure"
    wrapper.update_method = update_method

    return wrapper


class PlottingContextBase:
    """base class of the plotting contexts

    Example:
        The context wraps calls to the :mod:`matplotlib.pyplot` interface::

            context = PlottingContext()
            with context:
                plt.plot(...)
                plt.xlabel(...)
    """

    supports_update: bool = True
    """ flag indicating whether the context supports that plots can be updated
    with out redrawing the entire plot """

    def __init__(self, title: Optional[str] = None, show: bool = True):
        """
        Args:
            title (str): The shown in the plot
            show (bool): Flag determining whether plots are actually shown
        """
        self.title = title
        self.show = show

        self.initial_plot = True
        self.fig = None
        self._logger = logging.getLogger(__name__)
        self._logger.info(f"Initialize {self.__class__.__name__}")

    def __enter__(self):
        # start the plotting process
        if self.fig is not None:
            import matplotlib.pyplot as plt

            self.fig = plt.figure(self.fig.number)

    def __exit__(self, *exc):
        if self.initial_plot or not self.supports_update:
            # recreate the entire figure
            import matplotlib.pyplot as plt

            self.fig = plt.gcf()
            if len(self.fig.axes) == 0:
                # The figure seems to be empty, which must be a mistake
                raise RuntimeError("Plot figure does not contain axes")

            elif len(self.fig.axes) == 1:
                # The figure contains only a single axis, indicating that it is
                # composed of a single panel
                self._title = plt.title(self.title)

            else:
                # The figure contains multiple axes. This is an indication that
                # the figure consists of multiple panels, although insets and
                # colorbars also count as additional axes
                self._title = plt.suptitle(self.title)

            self.initial_plot = False

        else:
            # update the old figure
            self._title.set_text(self.title)

    def close(self):
        """close the plot"""
        # close matplotlib figure
        if self.fig is not None:
            import matplotlib.pyplot as plt

            plt.close(self.fig)


class BasicPlottingContext(PlottingContextBase):
    """basic plotting using just matplotlib"""

    def __init__(self, fig_or_ax=None, title: Optional[str] = None, show: bool = True):
        """
        Args:
            fig_or_ax:
                If axes are given, they are used. If a figure is given, it is
                set as active.
            title (str):
                The shown in the plot
            show (bool):
                Flag determining whether plots are actually shown
        """
        import matplotlib.axes as mpl_axes
        import matplotlib.figure as mpl_figure

        super().__init__(title=title, show=show)

        # determine which figure to modify
        if isinstance(fig_or_ax, mpl_axes.Axes):
            self.fig = fig_or_ax.get_figure()  # assume that axes are given
        elif isinstance(fig_or_ax, mpl_figure.Figure):
            self.fig = fig_or_ax

    def __exit__(self, *exc):
        super().__exit__(*exc)
        if self.show:
            self.fig.canvas.draw()  # required for display in nbagg backend
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                # add a small pause to allow the GUI to run it's event loop
                import matplotlib.pyplot as plt

                plt.pause(1e-3)


class JupyterPlottingContext(PlottingContextBase):
    """plotting in a jupyter widget using the `inline` backend"""

    supports_update = False
    """ flag indicating whether the context supports that plots can be updated
    with out redrawing the entire plot. The jupyter backend (`inline`) requires
    replotting of the entire figure, so an update is not supported."""

    def __enter__(self):
        from IPython.display import display
        from ipywidgets import Output

        if self.initial_plot:
            # close all previous plots
            import matplotlib.pyplot as plt

            plt.close("all")

            # create output widget for capturing all plotting
            self._ipython_out = Output()

            if self.show:
                # only show the widget if necessary
                display(self._ipython_out)

        # capture plots in the output widget
        self._ipython_out.__enter__()

    def __exit__(self, *exc):
        import matplotlib.pyplot as plt

        # finalize plot
        super().__exit__(*exc)

        if self.show:
            # show the plot, but ...
            plt.show()  # show the figure to make sure it can be captured
        # ... also clear it the next time something is done
        self._ipython_out.clear_output(wait=True)

        # stop capturing plots in the output widget
        self._ipython_out.__exit__(*exc)

        # close the figure, so figure windows do not accumulate
        plt.close(self.fig)

    def close(self):
        """close the plot"""
        super().close()
        # close ipython output
        try:
            self._ipython_out.close()
        except Exception:
            pass


def get_plotting_context(
    context=None, title: Optional[str] = None, show: bool = True
) -> PlottingContextBase:
    """returns a suitable plotting context

    Args:
        context:
            An instance of :class:`PlottingContextBase` or an instance of
            :class:`matplotlib.axes.Axes` or :class:`matplotlib.figure.Figure` to
            determine where the plotting will happen. If omitted, the context is
            determined automatically.
        title (str):
            The title shown in the plot
        show (bool):
            Determines whether the plot is shown while the simulation is running. If
            `False`, the files are created in the background.

    Returns:
        :class:`PlottingContextBase`: The plotting context
    """
    import matplotlib as mpl
    import matplotlib.axes as mpl_axes
    import matplotlib.figure as mpl_figure

    if context is None:
        # figure out whether plots are shown in jupyter notebook

        if "backend_inline" in mpl.get_backend():
            # special context to support the `inline` backend
            try:
                from IPython.display import display  # @UnusedImport
                from ipywidgets import Output  # @UnusedImport
            except ImportError:
                context_class: Type[PlottingContextBase] = BasicPlottingContext
            else:
                context_class = JupyterPlottingContext

        else:
            # standard context for all other backends
            context_class = BasicPlottingContext

        return context_class(title=title, show=show)

    elif isinstance(context, PlottingContextBase):
        # re-use an existing context
        context.title = title
        context.show = show
        return context

    elif isinstance(context, (mpl_axes.Axes, mpl_figure.Figure)):
        # create a basic context based on the given axes or figure
        return BasicPlottingContext(fig_or_ax=context, title=title, show=show)

    else:
        raise RuntimeError(f"Unknown plotting context `{context}`")


def in_ipython() -> bool:
    """try to detect whether we are in an ipython shell, e.g., a jupyter notebook"""
    ipy_module = sys.modules.get("IPython")
    if ipy_module:
        return bool(ipy_module.get_ipython())
    else:
        return False


@contextlib.contextmanager
def napari_viewer(
    grid: GridBase, run: Optional[bool] = None, close: bool = False, **kwargs
) -> Generator[napari.viewer.Viewer, None, None]:
    """creates an napari viewer for interactive plotting

    Args:
        grid (:class:`pde.grids.base.GridBase`):
            The grid defining the space
        run (bool):
            Whether to run the event loop of napari.
        close (bool):
            Whether to close the viewer immediately (e.g. for testing)
        **kwargs:
            Extra arguments are passed to :class:`napari.Viewer`
    """
    import napari  # @Reimport

    # initialize the viewer
    kwargs.setdefault("axis_labels", grid.axes)
    kwargs.setdefault("ndisplay", 3 if grid.num_axes >= 3 else 2)
    viewer = napari.Viewer(**kwargs)

    # allow the calling code to add content to the viewer
    yield viewer

    # start the napari's event loop if requested or if we're not in ipython
    if run is None:
        run = not in_ipython()
    if run:
        napari.run()

    # close the viewer if requested
    if close:
        warnings.warn("Closing napari does not work reliably and is thus disabled")
        # viewer.close()


def napari_add_layers(
    viewer: napari.viewer.Viewer, layers_data: Dict[str, Dict[str, Any]]
):
    """adds layers to a `napari <http://napari.org/>`__ viewer

    Args:
        viewer (:class:`napar    i.viewer.Viewer`):
            The napari application
        layers_data (dict):
            Data for all layers that will be added.
    """
    for name, layer_data in layers_data.items():
        layer_data.setdefault("name", name)
        layer_type = layer_data.pop("type")
        try:
            add_layer = getattr(viewer, f"add_{layer_type}")
        except AttributeError:
            raise RuntimeError(f"Unknown layer type: {layer_type}")
        else:
            add_layer(**layer_data)



"""
.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
"""

import numba as nb
import numpy as np
import pytest
from scipy import ndimage

from fixtures.fields import iter_grids
from pde.fields import FieldCollection, ScalarField, Tensor2Field, VectorField
from pde.fields.base import DataFieldBase, FieldBase
from pde.grids import (
    CartesianGrid,
    CylindricalSymGrid,
    PolarSymGrid,
    SphericalSymGrid,
    UnitGrid,
)
from pde.tools.misc import skipUnlessModule


@pytest.mark.parametrize("field_class", [ScalarField, VectorField, Tensor2Field])
def test_set_label(field_class):
    """test setting the field labels"""
    grid = UnitGrid([2])
    assert field_class(grid).label is None
    f = field_class(grid, label="a")
    assert f.label == "a"
    f.label = "b"
    assert f.label == "b"
    f.label = None
    assert f.label is None

    with pytest.raises(TypeError):
        f.label = 3
    with pytest.raises(TypeError):
        field_class(grid, label=1)


@pytest.mark.parametrize("grid", iter_grids())
@pytest.mark.parametrize("field_class", [ScalarField, Tensor2Field])
def test_interpolation_natural(grid, field_class, rng):
    """test some interpolation for natural boundary conditions"""
    msg = f"grid={grid}, field={field_class}"
    f = field_class.random_uniform(grid, rng=rng)

    def get_point():
        if isinstance(grid, CartesianGrid):
            return grid.get_random_point(boundary_distance=0.5, coords="grid", rng=rng)
        else:
            return grid.get_random_point(
                boundary_distance=1, avoid_center=True, coords="grid", rng=rng
            )

    # interpolate at cell center
    c = (1,) * len(grid.axes)
    p = f.grid.cell_coords[c]
    val = f.interpolate(p)
    np.testing.assert_allclose(val, f.data[(Ellipsis,) + c], err_msg=msg)


@pytest.mark.parametrize("num", [1, 3])
@pytest.mark.parametrize("grid", iter_grids())
def test_shapes_nfields(num, grid, rng):
    """test single component field"""
    fields = [ScalarField.random_uniform(grid, rng=rng) for _ in range(num)]
    field = FieldCollection(fields)
    data_shape = (num,) + grid.shape
    np.testing.assert_equal(field.data.shape, data_shape)
    for pf_single in field:
        np.testing.assert_equal(pf_single.data.shape, grid.shape)

    field_c = field.copy()
    np.testing.assert_allclose(field.data, field_c.data)
    assert field.grid == field_c.grid


@pytest.mark.parametrize("field_class", [ScalarField, VectorField, Tensor2Field])
def test_arithmetics(field_class):
    """test simple arithmetics for fields"""
    grid = UnitGrid([2, 2])
    f1 = field_class(grid, data=1)
    f2 = field_class(grid, data=2)
    assert isinstance(str(f1), str)
    np.testing.assert_allclose(f1.data, 1)

    np.testing.assert_allclose((-f1).data, -1)

    # test addition
    np.testing.assert_allclose((f1 + 1).data, 2)
    np.testing.assert_allclose((1 + f1).data, 2)
    f1 += 1
    np.testing.assert_allclose(f1.data, 2)
    np.testing.assert_allclose((f1 + f2).data, 4)

    # test subtraction
    np.testing.assert_allclose((f1 - 1).data, 1)
    np.testing.assert_allclose((1 - f1).data, -1)
    f1 -= 1
    np.testing.assert_allclose(f1.data, 1)
    np.testing.assert_allclose((f1 - f2).data, -1)

    # test multiplication
    np.testing.assert_allclose((f1 * 2).data, 2)
    np.testing.assert_allclose((2 * f1).data, 2)
    f1 *= 2
    np.testing.assert_allclose(f1.data, 2)

    # test division
    np.testing.assert_allclose((f1 / 2).data, 1)
    f1.data = 4
    np.testing.assert_allclose((2 / f1).data, 0.5)
    f1 /= 2
    np.testing.assert_allclose(f1.data, 2)

    # test power
    f1.data = 2
    np.testing.assert_allclose((f1**3).data, 8)
    f1 **= 3
    np.testing.assert_allclose(f1.data, 8)

    # test applying a function
    f1.data = 2
    np.testing.assert_allclose(f1.apply(lambda x: x**3).data, 8)
    np.testing.assert_allclose(f1.apply("c**3").data, 8)
    f1_out = f1.copy()
    f1.apply(lambda x: x**3, out=f1_out)
    np.testing.assert_allclose(f1_out.data, 8)
    f1_out = f1.copy()
    f1.apply("c**3", out=f1_out)
    np.testing.assert_allclose(f1_out.data, 8)


def test_scalar_arithmetics(rng):
    """test simple arithmetics involving scalar fields"""
    grid = UnitGrid([3, 4])
    s = ScalarField(grid, data=2)
    v = VectorField.random_uniform(grid, rng=rng)

    for f in [v, FieldCollection([v])]:
        f.data = s
        assert f.data.shape == (2, 3, 4)
        np.testing.assert_allclose(f.data, 2)

        f += s
        np.testing.assert_allclose(f.data, 4)
        np.testing.assert_allclose((f + s).data, 6)
        np.testing.assert_allclose((s + f).data, 6)
        f -= s
        np.testing.assert_allclose((f - s).data, 0)
        np.testing.assert_allclose((s - f).data, 0)

        f *= s
        np.testing.assert_allclose(f.data, 4)
        np.testing.assert_allclose((f * s).data, 8)
        np.testing.assert_allclose((s * f).data, 8)
        f /= s
        np.testing.assert_allclose((f / s).data, 1)
        with pytest.raises(TypeError):
            s / f
        with pytest.raises(TypeError):
            s /= f
        with pytest.raises(TypeError):
            s *= f


@pytest.mark.parametrize("field_class", [ScalarField, VectorField, Tensor2Field])
def test_data_managment(field_class):
    """test how data is set"""
    grid = UnitGrid([2, 2])
    s1 = field_class(grid, data=1)
    np.testing.assert_allclose(s1.data, 1)

    s2 = field_class(grid)
    np.testing.assert_allclose(s2.data, 0)

    c = FieldCollection([s1, s2])
    s1.data = 0
    np.testing.assert_allclose(c.data, 0)

    c.data = 2
    np.testing.assert_allclose(s1.data, 2)
    np.testing.assert_allclose(s2.data, 2)

    c.data += 1
    np.testing.assert_allclose(s1.data, 3)
    np.testing.assert_allclose(s2.data, 3)

    c[0].data += 2  # reference to s1
    c[1].data *= 2  # reference to s2
    np.testing.assert_allclose(s1.data, 5)
    np.testing.assert_allclose(s2.data, 6)

    c[0] = s2
    np.testing.assert_allclose(c.data, 6)

    # nested collections
    with pytest.raises(RuntimeError):
        FieldCollection([c])


@pytest.mark.parametrize("field_class", [ScalarField, VectorField, Tensor2Field])
def test_complex_fields(field_class, rng):
    """test operations on complex fields"""
    grid = UnitGrid([3])

    field = field_class.random_uniform(grid, 0, 1 + 1j, rng=rng)
    assert field.is_complex
    assert field.dtype == np.dtype("complex")

    field_copy = field.copy()
    assert field_copy.is_complex
    assert field_copy.dtype == np.dtype("complex")


@skipUnlessModule("h5py")
def test_hdf_input_output(tmp_path, rng):
    """test writing and reading files"""
    grid = UnitGrid([4, 4])
    s = ScalarField.random_uniform(grid, label="scalar", rng=rng)
    v = VectorField.random_uniform(grid, label="vector", rng=rng)
    t = Tensor2Field.random_uniform(grid, label="tensor", rng=rng)
    col = FieldCollection([s, v, t], label="collection")

    path = tmp_path / "test_hdf_input_output.hdf5"
    for f in [s, v, t, col]:
        f.to_file(path)
        f2 = FieldBase.from_file(path)
        assert f == f2
        assert f.label == f2.label
        assert isinstance(str(f), str)
        assert isinstance(repr(f), str)


def test_writing_images(tmp_path, rng):
    """test writing and reading files"""
    from matplotlib.pyplot import imread

    grid = UnitGrid([4, 4])
    s = ScalarField.random_uniform(grid, label="scalar", rng=rng)
    v = VectorField.random_uniform(grid, label="vector", rng=rng)
    t = Tensor2Field.random_uniform(grid, label="tensor", rng=rng)

    path = tmp_path / "test_writing_images.png"
    for f in [s, v, t]:
        f.to_file(path)
        # try reading the file
        with path.open("br") as fp:
            imread(fp)


@pytest.mark.parametrize("ndim", [1, 2])
def test_interpolation_to_grid_fields(ndim):
    """test whether data is interpolated correctly for different fields"""
    grid = CartesianGrid([[0, 2 * np.pi]] * ndim, 6)
    grid2 = CartesianGrid([[0, 2 * np.pi]] * ndim, 8)
    if ndim == 1:
        vf = VectorField.from_expression(grid, ["cos(x)"])
    elif ndim == 2:
        vf = VectorField.from_expression(grid, ["sin(y)", "cos(x)"])
    sf = vf[0]  # test extraction of fields
    fc = FieldCollection([sf, vf])

    for f in [sf, vf, fc]:
        # test self-interpolation
        f0 = f.interpolate_to_grid(grid)
        np.testing.assert_allclose(f.data, f0.data, atol=1e-15)

        # test interpolation to finer grid and back
        f2 = f.interpolate_to_grid(grid2)
        f3 = f2.interpolate_to_grid(grid)
        np.testing.assert_allclose(f.data, f3.data, atol=0.2, rtol=0.2)


@pytest.mark.parametrize("field_cls", [ScalarField, VectorField, Tensor2Field])
def test_interpolation_values(field_cls, rng):
    """test whether data is interpolated correctly for different fields"""
    grid = UnitGrid([3, 4])
    f = field_cls.random_uniform(grid, rng=rng)
    f.set_ghost_cells("auto_periodic_neumann")

    intp = f.make_interpolator()
    c = f.grid.cell_coords[2, 2]
    np.testing.assert_allclose(intp(c), f.data[..., 2, 2])

    intp = f.make_interpolator(with_ghost_cells=True)
    c = f.grid.cell_coords[2, 2]
    np.testing.assert_allclose(intp(c, f._data_full), f.data[..., 2, 2])

    with pytest.raises((ValueError, IndexError)):
        intp(np.array([100, -100]))

    res = f.make_interpolator(fill=45)(np.array([100, -100]))
    np.testing.assert_almost_equal(res, np.full(f.data_shape, 45))


def test_interpolation_ghost_cells():
    """test whether data is interpolated correctly with or without ghost cells"""
    grid = UnitGrid([3])
    f = ScalarField(grid, [1, 2, 3])
    f.set_ghost_cells({"value": 0})

    intp = f.make_interpolator()
    res = intp(np.c_[0:4])
    np.testing.assert_allclose(res, np.array([1.0, 1.5, 2.5, 3.0]))

    intp = f.make_interpolator(with_ghost_cells=True)
    res = intp(np.c_[0:4])
    np.testing.assert_allclose(res, np.array([0, 1.5, 2.5, 0]))

    grid = UnitGrid([3], periodic=True)
    f = ScalarField(grid, [1, 2, 3])
    f.set_ghost_cells("periodic")

    intp = f.make_interpolator()
    res = intp(np.c_[0:4])
    np.testing.assert_allclose(res, np.array([2.0, 1.5, 2.5, 2.0]))

    intp = f.make_interpolator(with_ghost_cells=True)
    res = intp(np.c_[0:4])
    np.testing.assert_allclose(res, np.array([2.0, 1.5, 2.5, 2.0]))


@pytest.mark.parametrize(
    "grid",
    [
        UnitGrid((6,)),
        PolarSymGrid(6, 4),
        SphericalSymGrid(7, 4),
        CylindricalSymGrid(6, (0, 8), (7, 8)),
    ],
)
def test_interpolation_to_cartesian(grid):
    """test whether data is interpolated correctly to Cartesian grid"""
    dim = grid.dim
    vf = VectorField(grid, 2)
    sf = vf[0]  # test extraction of fields
    fc = FieldCollection([sf, vf])

    # subset
    grid_cart = UnitGrid([4] * dim)
    for f in [sf, fc]:
        res = f.interpolate_to_grid(grid_cart)
        np.testing.assert_allclose(res.data, 2)

    # superset
    grid_cart = UnitGrid([8] * dim)
    for f in [sf, fc]:
        res = f.interpolate_to_grid(grid_cart, fill=0)
        assert res.data.min() == 0
        assert res.data.max() == pytest.approx(2)


@pytest.mark.parametrize(
    "grid",
    [PolarSymGrid(6, 4), SphericalSymGrid(7, 4), CylindricalSymGrid(6, (0, 8), (7, 8))],
)
def test_get_cartesian_grid(grid):
    """test whether Cartesian grids can be created"""
    cart = grid.get_cartesian_grid(mode="valid")
    assert cart.volume < grid.volume
    cart = grid.get_cartesian_grid(mode="full")
    assert cart.volume > grid.volume


@pytest.mark.parametrize("grid", iter_grids())
def test_simple_plotting(grid, rng):
    """test simple plotting of various fields on various grids"""
    vf = VectorField.random_uniform(grid, rng=rng)
    tf = Tensor2Field.random_uniform(grid, rng=rng)
    sf = tf[0, 0]  # test extraction of fields
    fc = FieldCollection([sf, vf])
    for f in [sf, vf, tf, fc]:
        f.plot(action="close")
        f.plot(kind="line", action="close")
        if grid.dim >= 2:
            f.plot(kind="image", action="close")
        if isinstance(f, VectorField) and grid.dim == 2:
            f.plot(kind="quiver", action="close")
            f.plot(kind="streamplot", action="close")


@pytest.mark.parametrize("field_cls", [ScalarField, VectorField, Tensor2Field])
def test_random_uniform(field_cls, rng):
    """test whether random uniform fields behave correctly"""
    grid = UnitGrid([256, 256])
    a = rng.random()
    b = 2 + rng.random()
    f = field_cls.random_uniform(grid, a, b, rng=rng)
    assert np.mean(f.average) == pytest.approx((a + b) / 2, rel=0.02)
    assert np.std(f.data) == pytest.approx(0.288675 * (b - a), rel=0.1)

    np.testing.assert_allclose(f.real.data, f.data)
    np.testing.assert_allclose(f.imag.data, 0)


def test_random_uniform_types(rng):
    """test whether random uniform fields behave correctly for different types"""
    grid = UnitGrid([8])
    for dtype in [bool, int, float, complex]:
        field = VectorField.random_uniform(grid, dtype=dtype, rng=rng)
        assert field.dtype == np.dtype(dtype)
        assert isinstance(field.data.flat[0].item(), dtype)

    assert ScalarField.random_uniform(grid, 0, 1, rng=rng).dtype == np.dtype(float)
    assert ScalarField.random_uniform(grid, vmin=0 + 0j, rng=rng).dtype == np.dtype(
        complex
    )
    assert ScalarField.random_uniform(grid, vmax=1 + 0j, rng=rng).dtype == np.dtype(
        complex
    )
    assert ScalarField.random_uniform(grid, 0 + 0j, 1 + 0j, rng=rng).dtype == np.dtype(
        complex
    )


@pytest.mark.parametrize("field_cls", [ScalarField, VectorField, Tensor2Field])
def test_random_normal(field_cls, rng):
    """test whether random normal fields behave correctly"""
    grid = UnitGrid([256, 256])
    m = rng.random()
    s = 1 + rng.random()
    for scaling in ["none", "physical"]:
        f = field_cls.random_normal(grid, mean=m, std=s, scaling=scaling, rng=rng)
        assert np.mean(f.average) == pytest.approx(m, rel=0.1, abs=0.1)
        assert np.std(f.data) == pytest.approx(s, rel=0.1, abs=0.1)


def test_random_normal_types(rng):
    """test whether random normal fields behave correctly for different types"""
    grid = UnitGrid([8])
    for dtype in [bool, int, float, complex]:
        field = VectorField.random_normal(grid, dtype=dtype, rng=rng)
        assert field.dtype == np.dtype(dtype)
        assert isinstance(field.data.flat[0].item(), dtype)

    assert ScalarField.random_normal(grid, 0, 1, rng=rng).dtype == np.dtype(float)
    assert ScalarField.random_normal(grid, mean=0 + 0j, rng=rng).dtype == np.dtype(
        complex
    )
    assert ScalarField.random_normal(grid, std=1 + 0j, rng=rng).dtype == np.dtype(
        complex
    )
    assert ScalarField.random_normal(grid, 0 + 0j, 1 + 0j, rng=rng).dtype == np.dtype(
        complex
    )

    m = complex(rng.random(), rng.random())
    s = complex(1 + rng.random(), 1 + rng.random())
    grid = UnitGrid([256, 256])
    field = field.random_normal(grid, m, s, rng=rng)
    assert np.mean(field.average) == pytest.approx(m, rel=0.1, abs=0.1)
    assert np.std(field.data.real) == pytest.approx(s.real, rel=0.1, abs=0.1)
    assert np.std(field.data.imag) == pytest.approx(s.imag, rel=0.1, abs=0.1)


@pytest.mark.parametrize("field_cls", [ScalarField, VectorField, Tensor2Field])
def test_random_colored(field_cls, rng):
    """test whether random colored fields behave correctly"""
    grid = UnitGrid([128, 128])
    exponent = rng.uniform(-4, 4)
    scale = 1 + rng.random()
    f = field_cls.random_colored(grid, exponent=exponent, scale=scale, rng=rng)

    assert np.allclose(f.average, 0)


def test_random_rng():
    """test whether the random number generator arguments are accepted"""
    grid = UnitGrid([2, 2])
    for create_random_field in [
        ScalarField.random_colored,
        ScalarField.random_harmonic,
        ScalarField.random_normal,
        ScalarField.random_uniform,
    ]:
        f1 = create_random_field(grid, rng=np.random.default_rng(0))
        f2 = create_random_field(grid, rng=np.random.default_rng(0))
        np.testing.assert_allclose(f1.data, f2.data)


@pytest.mark.parametrize("dim", [1, 2])
@pytest.mark.parametrize("size", [256, 512])
def test_fluctuations(dim, size, rng):
    """test the scaling of fluctuations"""
    if dim == 1:
        size **= 2
    grid = CartesianGrid([[0, 1]] * dim, [size] * dim)
    std = 1 + rng.random()
    for field_cls in [ScalarField, VectorField, Tensor2Field]:
        s = field_cls.random_normal(
            grid, mean=rng.random(), std=std, scaling="physical", rng=rng
        )
        expect = np.full([dim] * field_cls.rank, std)
        np.testing.assert_allclose(s.fluctuations, expect, rtol=0.1)


def test_smoothing(rng):
    """test smoothing on different grids"""
    for grid in [
        CartesianGrid([[-2, 3]], 4),
        UnitGrid(7, periodic=False),
        UnitGrid(7, periodic=True),
    ]:
        f1 = ScalarField.random_uniform(grid, rng=rng)
        sigma = 0.5 + rng.random()

        # this assumes that the grid periodicity is the same for all axes
        mode = "wrap" if grid.periodic[0] else "reflect"
        s = sigma / grid.typical_discretization
        expected = ndimage.gaussian_filter(f1.data, sigma=s, mode=mode)

        out = f1.smooth(sigma)
        np.testing.assert_allclose(out.data, expected)

        out.data = 0  # reset data
        f1.smooth(sigma, out=out).data
        np.testing.assert_allclose(out.data, expected)

    # test one simple higher order smoothing
    tf = Tensor2Field.random_uniform(grid, rng=rng)
    assert tf.data.shape == tf.smooth(1).data.shape

    # test in-place smoothing
    g = UnitGrid([8, 8])
    f1 = ScalarField.random_normal(g, rng=rng)
    f2 = f1.smooth(3)
    f1.smooth(3, out=f1)
    np.testing.assert_allclose(f1.data, f2.data)


def test_vector_from_scalars():
    """test how to compile vector fields from scalar fields"""
    g = UnitGrid([1, 2])
    s1 = ScalarField(g, [[0, 1]])
    s2 = ScalarField(g, [[2, 3]])
    v = VectorField.from_scalars([s1, s2], label="test")
    assert v.label == "test"
    np.testing.assert_equal(v.data, [[[0, 1]], [[2, 3]]])

    with pytest.raises(ValueError):
        VectorField.from_scalars([s1, s2, s1])


@pytest.mark.parametrize(
    "grid", [UnitGrid([3, 2]), UnitGrid([3]), CylindricalSymGrid(1, (0, 2), 3)]
)
def test_dot_product(grid, rng):
    """test dot products between vectors and tensors"""
    vf = VectorField.random_normal(grid, rng=rng)
    tf = Tensor2Field.random_normal(grid, rng=rng)
    dot = vf.make_dot_operator()

    expected = np.einsum("i...,i...->...", vf.data, vf.data)
    np.testing.assert_allclose((vf @ vf).data, expected)
    np.testing.assert_allclose(dot(vf.data, vf.data), expected)

    expected = np.einsum("i...,i...->...", vf.data, tf.data)
    np.testing.assert_allclose((vf @ tf).data, expected)
    np.testing.assert_allclose(dot(vf.data, tf.data), expected)

    expected = np.einsum("ji...,i...->j...", tf.data, vf.data)
    np.testing.assert_allclose((tf @ vf).data, expected)
    np.testing.assert_allclose(dot(tf.data, vf.data), expected)

    expected = np.einsum("ij...,jk...->ik...", tf.data, tf.data)
    np.testing.assert_allclose((tf @ tf).data, expected)
    np.testing.assert_allclose(dot(tf.data, tf.data), expected)

    # test non-sensical dot product
    sf = ScalarField.random_normal(grid, rng=rng)
    # vector dot
    with pytest.raises(TypeError):
        vf @ sf
    with pytest.raises(TypeError):
        tf @ sf
    with pytest.raises((TypeError, nb.errors.TypingError)):
        dot(vf.data, sf.data)
    with pytest.raises((TypeError, nb.errors.TypingError)):
        dot(sf.data, vf.data)
    with pytest.raises((TypeError, nb.errors.TypingError)):
        dot(tf.data, sf.data)
    with pytest.raises((TypeError, nb.errors.TypingError)):
        dot(sf.data, tf.data)


@pytest.mark.parametrize("grid", iter_grids())
def test_complex_operator(grid, rng):
    """test using a complex operator on grid"""
    r = ScalarField.random_normal(grid, rng=rng)
    i = ScalarField.random_normal(grid, rng=rng)
    c = r + 1j * i
    assert c.is_complex
    assert np.iscomplexobj(c)

    c_lap = c.laplace("auto_periodic_neumann").data
    np.testing.assert_allclose(c_lap.real, r.laplace("auto_periodic_neumann").data)
    np.testing.assert_allclose(c_lap.imag, i.laplace("auto_periodic_neumann").data)


def test_get_field_class_by_rank():
    """test _get_field_class_by_rank function"""
    assert DataFieldBase.get_class_by_rank(0) is ScalarField
    assert DataFieldBase.get_class_by_rank(1) is VectorField
    assert DataFieldBase.get_class_by_rank(2) is Tensor2Field
    with pytest.raises(RuntimeError):
        DataFieldBase.get_class_by_rank(3)



"""
Defines a scalar field over a grid

.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
"""

from __future__ import annotations

import numbers
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Callable,
    Dict,
    List,
    Literal,
    Optional,
    Sequence,
    Tuple,
    Union,
)

import numpy as np
from numpy.typing import DTypeLike

from ..grids import CartesianGrid, UnitGrid
from ..grids.base import DomainError, GridBase
from ..tools.docstrings import fill_in_docstring
from ..tools.misc import Number
from ..tools.typing import NumberOrArray
from .base import DataFieldBase

if TYPE_CHECKING:
    from ..grids.boundaries.axes import BoundariesData  # @UnusedImport
    from .vectorial import VectorField  # @UnusedImport


class ScalarField(DataFieldBase):
    """Scalar field discretized on a grid"""

    rank = 0

    @classmethod
    @fill_in_docstring
    def from_expression(
        cls,
        grid: GridBase,
        expression: str,
        *,
        user_funcs: Optional[Dict[str, Callable]] = None,
        consts: Optional[Dict[str, NumberOrArray]] = None,
        label: Optional[str] = None,
        dtype: Optional[DTypeLike] = None,
    ) -> ScalarField:
        """create a scalar field on a grid from a given expression

        Warning:
            {WARNING_EXEC}

        Args:
            grid (:class:`~pde.grids.base.GridBase`):
                Grid defining the space on which this field is defined
            expression (str):
                Mathematical expression for the scalar value as a function of the
                position on the grid. The expression may contain standard mathematical
                functions and it may depend on the axes labels of the grid.
                More information can be found in the
                :ref:`expression documentation <documentation-expressions>`.
            user_funcs (dict, optional):
                A dictionary with user defined functions that can be used in the
                expression
            consts (dict, optional):
                A dictionary with user defined constants that can be used in the
                expression. The values of these constants should either be numbers or
                :class:`~numpy.ndarray`.
            label (str, optional):
                Name of the field
            dtype (numpy dtype):
                The data type of the field. If omitted, it will be determined from
                `data` automatically.
        """
        from ..tools.expressions import ScalarExpression

        # parse the expression
        expr = ScalarExpression(
            expression=expression,
            signature=grid.axes,
            user_funcs=user_funcs,
            consts=consts,
        )
        # obtain the coordinates of the grid points
        points = [grid.cell_coords[..., i] for i in range(grid.num_axes)]

        try:
            # try evaluating the expression using a vectorized call
            data = expr(*points)
        except ValueError:
            # if this fails, evaluate expression point-wise
            data = np.empty(grid.shape)
            for cells in np.ndindex(*grid.shape):
                data[cells] = expr(*grid.cell_coords[cells])

        return cls(grid=grid, data=data, label=label, dtype=dtype)

    @classmethod
    def from_image(
        cls,
        path: Union[Path, str],
        bounds=None,
        periodic=False,
        *,
        label: Optional[str] = None,
    ) -> ScalarField:
        """create a scalar field from an image

        Args:
            path (:class:`Path` or str):
                The path to the image file
            bounds (tuple, optional):
                Gives the coordinate range for each axis. This should be two tuples of
                two numbers each, which mark the lower and upper bound for each axis.
            periodic (bool or list):
                Specifies which axes possess periodic boundary conditions. This is
                either a list of booleans defining periodicity for each individual axis
                or a single boolean value specifying the same periodicity for all axes.
            label (str, optional):
                Name of the field
        """
        from matplotlib.pyplot import imread

        # read image and convert to grayscale
        data = imread(str(path))
        if data.ndim == 2:
            pass  # is already gray scale
        elif data.ndim == 3:
            # convert to gray scale using ITU-R 601-2 luma transform:
            weights = np.array([0.299, 0.587, 0.114])
            data = data[..., :3] @ weights
        else:
            raise RuntimeError(f"Image data has wrong shape: {data.shape}")

        # transpose data to use mathematical conventions for axes
        data = data.T[:, ::-1]

        # determine the associated grid
        if bounds is None:
            grid: GridBase = UnitGrid(data.shape, periodic=periodic)
        else:
            grid = CartesianGrid(bounds, data.shape, periodic=periodic)

        return cls(grid, data, label=label)

    @DataFieldBase._data_flat.setter  # type: ignore
    def _data_flat(self, value):
        """set the data from a value from a collection"""
        self._data_full = value[0]

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        """support unary numpy ufuncs, like np.sin, but also np.multiply"""
        if method == "__call__":
            # only support unary functions in simple calls

            # check the input
            arrs = []
            for arg in inputs:
                if isinstance(arg, numbers.Number):
                    arrs.append(arg)
                elif isinstance(arg, np.ndarray):
                    if arg.shape != self.data.shape:
                        raise RuntimeError(
                            f"Data shapes incompatible ({arg.shape} != {self.data.shape}"
                        )
                    arrs.append(arg)
                elif isinstance(arg, self.__class__):
                    self.assert_field_compatible(arg)
                    arrs.append(arg.data)
                else:
                    # unsupported type
                    return NotImplemented

            if "out" in kwargs:
                # write to given field
                out = kwargs.pop("out")[0]
                self.assert_field_compatible(out)
                kwargs["out"] = (out.data,)
                ufunc(*arrs, **kwargs)
                return out
            else:
                # return new field
                return self.__class__(self.grid, data=ufunc(*arrs, **kwargs))
        else:
            return NotImplemented

    @fill_in_docstring
    def laplace(
        self,
        bc: Optional[BoundariesData],
        out: Optional[ScalarField] = None,
        **kwargs,
    ) -> ScalarField:
        """apply Laplace operator and return result as a field

        Args:
            bc:
                The boundary conditions applied to the field.
                {ARG_BOUNDARIES_OPTIONAL}
            out (ScalarField, optional):
                Optional scalar field to which the  result is written.
            label (str, optional):
                Name of the returned field
            backend (str):
                The backend (e.g., 'numba' or 'scipy') used for this operator.

        Returns:
            :class:`~pde.fields.scalar.ScalarField`: the Laplacian of the field
        """
        return self.apply_operator("laplace", bc=bc, out=out, **kwargs)  # type: ignore

    @fill_in_docstring
    def gradient_squared(
        self,
        bc: Optional[BoundariesData],
        out: Optional[ScalarField] = None,
        **kwargs,
    ) -> ScalarField:
        r"""apply squared gradient operator and return result as a field

        This evaluates :math:`|\nabla \phi|^2` for the scalar field :math:`\phi`

        Args:
            bc:
                The boundary conditions applied to the field.
                {ARG_BOUNDARIES_OPTIONAL}
            out (ScalarField, optional):
                Optional vector field to which the result is written.
            label (str, optional):
                Name of the returned field
            central (bool):
                Determines whether a central difference approximation is used for the
                gradient operator or not. If not, the squared gradient is calculated as
                the mean of the squared values of the forward and backward derivatives,
                which thus includes the value at a support point in the result at the
                same point.

        Returns:
            :class:`~pde.fields.scalar.ScalarField`: the squared gradient of the field
        """
        return self.apply_operator("gradient_squared", bc=bc, out=out, **kwargs)  # type: ignore

    @fill_in_docstring
    def gradient(
        self,
        bc: Optional[BoundariesData],
        out: Optional["VectorField"] = None,
        **kwargs,
    ) -> "VectorField":
        """apply gradient operator and return result as a field

        Args:
            bc:
                The boundary conditions applied to the field.
                {ARG_BOUNDARIES_OPTIONAL}
            out (VectorField, optional):
                Optional vector field to which the result is written.
            label (str, optional):
                Name of the returned field

        Returns:
            :class:`~pde.fields.vectorial.VectorField`: result of applying the operator
        """
        return self.apply_operator("gradient", bc=bc, out=out, **kwargs)  # type: ignore

    @property
    def integral(self) -> Number:
        """Number: integral of the scalar field over space"""
        return self.grid.integrate(self.data)  # type: ignore

    def project(
        self,
        axes: Union[str, Sequence[str]],
        method: Literal["integral", "average", "mean"] = "integral",
        label: Optional[str] = None,
    ) -> ScalarField:
        """project scalar field along given axes

        Args:
            axes (list of str):
                The names of the axes that are removed by the projection
                operation. The valid names for a given grid are the ones in
                the :attr:`GridBase.axes` attribute.
            method (str):
                The projection method. This can be either 'integral' to
                integrate over the removed axes or 'average' to perform an
                average instead.
            label (str, optional):
                The label of the returned field

        Returns:
            :class:`~pde.fields.scalar.ScalarField`: The projected data in a scalar
            field with a subgrid of the original grid.
        """
        if isinstance(axes, str):
            axes = [axes]
        if any(ax not in self.grid.axes for ax in axes):
            raise ValueError(
                f"The axes {axes} are not all contained in {self.grid} with axes "
                f"{self.grid.axes}"
            )

        # determine the axes after projection
        ax_all = range(self.grid.num_axes)
        ax_remove = tuple(self.grid.axes.index(ax) for ax in axes)
        ax_retain = tuple(sorted(set(ax_all) - set(ax_remove)))

        # determine the new grid
        sliced_grid = self.grid.slice(ax_retain)

        # calculate the new data
        if method == "integral":
            subdata = self.grid.integrate(self.data, axes=ax_remove)

        elif method == "average" or method == "mean":
            integrals = self.grid.integrate(self.data, axes=ax_remove)
            volumes = self.grid.integrate(1, axes=ax_remove)
            subdata = integrals / volumes

        else:
            raise ValueError(f"Unknown projection method `{method}`")

        # create the new field instance
        return self.__class__(grid=sliced_grid, data=subdata, label=label)

    def slice(
        self,
        position: Dict[str, float],
        *,
        method: Literal["nearest"] = "nearest",
        label: Optional[str] = None,
    ) -> ScalarField:
        """slice data at a given position

        Note:
            This method should not be used to evaluate fields right at the boundary
            since it does not respect boundary conditions. Use
            :meth:`~ScalarField.get_boundary_field` to obtain the values directly on the
            boundary.

        Args:
            position (dict):
                Determines the location of the slice using a dictionary supplying
                coordinate values for a subset of axes. Axes not mentioned in the
                dictionary are retained and form the slice. For instance, in a 2d
                Cartesian grid, `position = {'x': 1}` slices along the y-direction at
                x=1. Additionally, the special positions 'low', 'mid', and 'high' are
                supported to reference relative positions along the axis.
            method (str):
                The method used for slicing. Currently, we only support `nearest`, which
                takes data from cells defined on the grid.
            label (str, optional):
                The label of the returned field

        Returns:
            :class:`~pde.fields.scalar.ScalarField`: The sliced data in a scalar field
            with a subgrid of the original grid.
        """
        grid = self.grid

        # parse the positions and determine the axes to remove
        ax_remove, pos_values = [], np.zeros(grid.num_axes)
        for ax, pos in position.items():
            # check the axis
            try:
                i = grid.axes.index(ax)
            except ValueError:
                raise ValueError(
                    f"The axes {ax} is not contained in "
                    f"{self.grid} with axes {self.grid.axes}"
                )
            ax_remove.append(i)

            # check the position
            if isinstance(pos, str):
                if pos in {"min", "low", "lower"}:
                    pos_values[i] = grid.axes_coords[i][0]
                elif pos in {"max", "high", "upper"}:
                    pos_values[i] = grid.axes_coords[i][-1]
                elif pos in {"mid", "middle", "center"}:
                    pos_values[i] = np.mean(grid.axes_bounds[i])
                else:
                    raise ValueError(f"Unknown position `{pos}`")
            else:
                pos_values[i] = float(pos)

        # determine the axes left after slicing and the new grid
        ax_all = range(grid.num_axes)
        ax_retain = tuple(sorted(set(ax_all) - set(ax_remove)))
        sliced_grid = grid.slice(ax_retain)

        # obtain the sliced data
        if method == "nearest":
            idx: List[Union[int, slice]] = []
            for i in range(grid.num_axes):
                if i in ax_remove:
                    pos = pos_values[i]
                    axis_bounds = grid.axes_bounds[i]
                    if pos < axis_bounds[0] or pos > axis_bounds[1]:
                        raise DomainError(
                            f"Position {grid.axes[i]} = {pos} is outside the domain"
                        )
                    # add slice that is closest to pos
                    idx.append(int(np.argmin((grid.axes_coords[i] - pos) ** 2)))
                else:
                    idx.append(slice(None))
            subdata = self.data[tuple(idx)]

        else:
            raise ValueError(f"Unknown slicing method `{method}`")

        # create the new field instance
        return self.__class__(grid=sliced_grid, data=subdata, label=label)

    def to_scalar(
        self, scalar: Union[str, Callable] = "auto", *, label: Optional[str] = None
    ) -> ScalarField:
        """return a modified scalar field by applying method `scalar`

        Args:
            scalar (str or callable):
                Determines the method used for obtaining the scalar. If this is a
                callable, it is simply applied to self.data and a new scalar field with
                this data is returned. Alternatively, pre-defined methods can be
                selected using strings. Here, `abs` and `norm` denote the norm of each
                entry of the field, while `norm_squared` returns the squared norm. The
                default `auto` is to return a (unchanged) copy of a real field and the
                norm of a complex field.
            label (str, optional):
                Name of the returned field

        Returns:
            :class:`~pde.fields.scalar.ScalarField`: Scalar field after applying the
            operation
        """
        if callable(scalar):
            data = scalar(self.data)
        elif scalar == "auto":
            if np.iscomplexobj(self.data):
                data = np.abs(self.data)
            else:
                data = self.data
        elif scalar == "abs" or scalar == "norm":
            data = np.abs(self.data)
        elif scalar == "norm_squared":
            data = self.data * self.data.conjugate()
        else:
            raise ValueError(f"Unknown method `{scalar}` for `to_scalar`")

        return ScalarField(grid=self.grid, data=data, label=label)

    @fill_in_docstring
    def get_boundary_field(
        self,
        index: Union[str, Tuple[int, bool]],
        bc: Optional[BoundariesData] = None,
        *,
        label: Optional[str] = None,
    ) -> ScalarField:
        """get the field on the specified boundary

        Args:
            index (str or tuple):
                Index specifying the boundary. Can be either a string given in
                :attr:`~pde.grids.base.GridBase.boundary_names`, like :code:`"left"`, or
                a tuple of the axis index perpendicular to the boundary and a boolean
                specifying whether the boundary is at the upper side of the axis or not,
                e.g., :code:`(1, True)`.
            bc:
                The boundary conditions applied to the field.
                {ARG_BOUNDARIES_OPTIONAL}
            label (str):
                Label of the returned field

        Returns:
            :class:`~pde.fields.scalar.ScalarField`: The field on the boundary
        """
        axis, upper = self.grid._get_boundary_index(index)
        data = self.get_boundary_values(axis, upper, bc)

        boundary_axes = tuple(i for i in range(self.grid.num_axes) if i != axis)
        if boundary_axes:
            # the boundary is an actual field (the original grid had more than 2 axes)
            grid = self.grid.slice(boundary_axes)
        else:
            # the boundary is a singular point => return a UnitGrid
            grid = UnitGrid([1])

        return self.__class__(grid=grid, data=data, label=label, dtype=self.dtype)



"""
Base classes for storing data
   
.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de> 
"""

from __future__ import annotations

import logging
from abc import ABCMeta, abstractmethod
from inspect import signature
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Iterator,
    List,
    Literal,
    Optional,
    Sequence,
    Tuple,
    Union,
)

import numpy as np
from numpy.typing import DTypeLike

from ..fields import FieldCollection, ScalarField, Tensor2Field, VectorField
from ..fields.base import FieldBase
from ..grids.base import GridBase
from ..tools.docstrings import fill_in_docstring
from ..tools.output import display_progress
from ..trackers.base import InfoDict, TrackerBase
from ..trackers.interrupts import InterruptsBase, IntervalData

if TYPE_CHECKING:
    from .memory import MemoryStorage  # @UnusedImport


WriteModeType = Literal[
    "append",
    "readonly" "truncate",
    "truncate_once",
]


class StorageBase(metaclass=ABCMeta):
    """base class for storing time series of discretized fields

    These classes store time series of :class:`~pde.fields.base.FieldBase`, i.e., they
    store the values of the fields at particular time points. Iterating of the storage
    will return the fields in order and individual time points can also be accessed.
    """

    times: Sequence[float]  # :class:`~numpy.ndarray`): stored time points
    data: Any  # actual data for all the stored times
    write_mode: WriteModeType  # mode determining how the storage behaves

    def __init__(
        self,
        info: Optional[InfoDict] = None,
        write_mode: WriteModeType = "truncate_once",
    ):
        """
        Args:
            info (dict):
                Supplies extra information that is stored in the storage
            write_mode (str):
                Determines how new data is added to already existing one. Possible
                values are: 'append' (data is always appended), 'truncate' (data is
                cleared every time this storage is used for writing), or 'truncate_once'
                (data is cleared for the first writing, but subsequent data using the
                same instances are appended). Alternatively, specifying 'readonly' will
                disable writing completely.
        """
        self.info = {} if info is None else info
        self.write_mode = write_mode
        self._data_shape: Optional[Tuple[int, ...]] = None
        self._dtype: Optional[DTypeLike] = None
        self._grid: Optional[GridBase] = None
        self._field: Optional[FieldBase] = None
        self._logger = logging.getLogger(self.__class__.__name__)

    @property
    def data_shape(self) -> Tuple[int, ...]:
        """the current data shape.

        Raises:
            RuntimeError: if data_shape was not set
        """
        if self._data_shape is None:
            raise RuntimeError("data_shape was not set")
        else:  # use the else clause to help typing
            return self._data_shape

    @property
    def dtype(self) -> DTypeLike:
        """the current data type.

        Raises:
            RuntimeError: if data_type was not set
        """
        if self._dtype is None:
            raise RuntimeError("dtype was not set")
        else:  # use the else clause to help typing
            return self._dtype

    @abstractmethod
    def _append_data(self, data: np.ndarray, time: float) -> None:
        pass

    def append(self, field: FieldBase, time: Optional[float] = None) -> None:
        """add field to the storage

        Args:
            field (:class:`~pde.fields.base.FieldBase`):
                The field that is added to the storage
            time (float, optional):
                The time point
        """
        if time is None:
            time = 0 if len(self) == 0 else self.times[-1] + 1

        if self._grid is None:
            self._grid = field.grid
        elif self._grid != field.grid:
            raise ValueError(f"Grids incompatible ({self._grid} != {field.grid})")
        return self._append_data(field.data, time)

    def clear(self, clear_data_shape: bool = False) -> None:
        """truncate the storage by removing all stored data.

        Args:
            clear_data_shape (bool):
                Flag determining whether the data shape is also deleted.
        """
        if clear_data_shape:
            self._data_shape = None
            self._dtype = None

    def __len__(self):
        """return the number of stored items, i.e., time steps"""
        return len(self.times)

    @property
    def shape(self) -> Optional[Tuple[int, ...]]:
        """the shape of the stored data"""
        if self._data_shape:
            return (len(self),) + self._data_shape
        else:
            return None

    @property
    def has_collection(self) -> bool:
        """bool: whether the storage is storing a collection"""
        if self._field is not None:
            return isinstance(self._field, FieldCollection)
        elif len(self) > 0:
            return isinstance(self._get_field(0), FieldCollection)
        else:
            raise RuntimeError("Storage is empty")

    @property
    def grid(self) -> Optional[GridBase]:
        """GridBase: the grid associated with this storage

        This returns `None` if grid was not stored in `self.info`.
        """
        if self._grid is None:
            if "field_attributes" in self.info:
                attrs_serialized = self.info["field_attributes"]
                attrs = FieldBase.unserialize_attributes(attrs_serialized)

                # load the grid if possible
                if "grid" in attrs:
                    # load grid from only stored field
                    self._grid = attrs["grid"]
                elif "fields" in attrs:
                    # load grid from first field of a collection
                    self._grid = attrs["fields"][0]["grid"]
                else:
                    self._logger.warning(
                        "`grid` attribute was not stored. Available attributes: "
                        + ", ".join(sorted(attrs.keys()))
                    )

            else:
                self._logger.warning("Field attributes are unavailable in info")
        return self._grid

    def _init_field(self) -> None:
        """initialize internal field variable"""
        if self.grid is None:
            raise RuntimeError(
                "Could not load grid from data. Please set the `_grid` attribute "
                "to the grid that has been used for the stored data"
            )

        if "field_attributes" in self.info:
            # field type was stored in data
            attrs_serialized = self.info["field_attributes"]
            attrs = FieldBase.unserialize_attributes(attrs_serialized)
            self._field = FieldBase.from_state(attrs)

        else:
            # try to determine field type automatically

            # obtain data shape by removing the first axis (associated with the time
            # series and the last axes (associated with the spatial dimensions). What is
            # left should be the (local) data stored at each grid point for each time
            # step. Note that self.data might be a list of arrays
            local_shape = self.data[0].shape[: -self.grid.num_axes]
            dim = self.grid.dim
            if len(local_shape) == 0:  # rank 0
                self._field = ScalarField(self.grid, dtype=self.data[0].dtype)
            elif local_shape == (dim,):  # rank 1
                self._field = VectorField(self.grid, dtype=self.data[0].dtype)
            elif local_shape == (dim, dim):  # rank 2
                self._field = Tensor2Field(self.grid, dtype=self.data[0].dtype)
            else:
                raise RuntimeError(
                    "`field` attribute was not stored in file and the data shape "
                    f"{local_shape} could not be interpreted automatically"
                )
            self._logger.warning(
                "`field` attribute was not stored. We guessed that the data is of "
                f"type {self._field.__class__.__name__}."
            )

    def _get_field(self, t_index: int) -> FieldBase:
        """return the field corresponding to the given time index

        Load the data given an index, i.e., the data at time `self.times[t_index]`.

        Args:
            t_index (int):
                The index of the data to load

        Returns:
            :class:`~pde.fields.FieldBase`:
            The field class containing the grid and data
        """
        if t_index < 0:
            t_index += len(self)

        if not 0 <= t_index < len(self):
            raise IndexError("Time index out of range")

        if self._field is None:
            self._init_field()

        # create the field with the data of the given index
        assert self._field is not None
        field = self._field.copy()
        field.data = self.data[t_index]
        return field

    def __getitem__(self, key: Union[int, slice]) -> Union[FieldBase, List[FieldBase]]:
        """return field at given index or a list of fields for a slice"""
        if isinstance(key, int):
            return self._get_field(key)
        elif isinstance(key, slice):
            return [self._get_field(i) for i in range(*key.indices(len(self)))]
        else:
            raise TypeError("Unknown key type")

    def __iter__(self) -> Iterator[FieldBase]:
        """iterate over all stored fields"""
        for i in range(len(self)):
            yield self[i]  # type: ignore

    def items(self) -> Iterator[Tuple[float, FieldBase]]:
        """iterate over all times and stored fields, returning pairs"""
        for i in range(len(self)):
            yield self.times[i], self[i]  # type: ignore

    @fill_in_docstring
    def tracker(
        self,
        interval: Union[int, float, InterruptsBase] = 1,
        *,
        transformation: Optional[Callable[[FieldBase, float], FieldBase]] = None,
    ) -> "StorageTracker":
        """create object that can be used as a tracker to fill this storage

        Args:
            interval:
                {ARG_TRACKER_INTERVAL}
            transformation (callable, optional):
                A function that transforms the current state into a new field or field
                collection, which is then stored. This allows to store derived
                quantities of the field during calculations. The argument needs to be a
                callable function taking 1 or 2 arguments. The first argument always is
                the current field, while the optional second argument is the associated
                time.

        Returns:
            :class:`StorageTracker`: The tracker that fills the current storage

        Example:
            The `transformation` argument allows storing additional fields:

            .. code-block:: python

                def add_to_state(state):
                    transformed_field = state.smooth(1)
                    return field.append(transformed_field)

                storage = pde.MemoryStorage()
                tracker = storage.tracker(1, transformation=add_to_state)
                eq.solve(..., tracker=tracker)

            In this example, :obj:`storage` will contain a trajectory of the fields of
            the simulation as well as the smoothed fields. Other transformations are
            possible by defining appropriate :func:`add_to_state`
        """
        return StorageTracker(
            storage=self, interval=interval, transformation=transformation
        )

    def start_writing(self, field: FieldBase, info: Optional[InfoDict] = None) -> None:
        """initialize the storage for writing data

        Args:
            field (:class:`~pde.fields.FieldBase`):
                An example of the data that will be written to extract the grid
                and the data_shape
            info (dict):
                Supplies extra information that is stored in the storage
        """
        if self.write_mode == "readonly":
            raise RuntimeError("Cannot write data in readonly mode")

        if self._data_shape is None:
            self._data_shape = field.data.shape
        elif self.data_shape != field.data.shape:
            raise ValueError("Data shape incompatible with stored data")

        if self._dtype is None:
            self._dtype = field.dtype

        self._grid = field.grid
        self._field = field.copy()
        self.info["field_attributes"] = field.attributes_serialized

    def end_writing(self) -> None:
        """finalize the storage after writing"""
        pass

    def extract_field(
        self, field_id: Union[int, str], label: Optional[str] = None
    ) -> "MemoryStorage":
        """extract the time course of a single field from a collection

        Note:
            This might return a view into the original data, so modifying the returned
            data can also change the underlying original data.

        Args:
            field_id (int or str):
                The index into the field collection. This determines which field of the
                collection is returned. Instead of a numerical index, the field label
                can also be supplied. If there are multiple fields with the same label,
                only the first field is returned.
            label (str):
                The label of the returned field. If omitted, the stored label is used.

        Returns:
            :class:`MemoryStorage`: a storage instance that contains the data for the
            single field
        """
        from .memory import MemoryStorage  # @Reimport

        if self._field is None:
            self._init_field()

        # get the field to check its type
        if not isinstance(self._field, FieldCollection):
            raise TypeError(
                "Can only extract fields from `FieldCollection`. Current storage "
                f"stores `{self._field.__class__.__name__}`."
            )

        # determine the field index
        if isinstance(field_id, str):
            field_index = self._field.labels.index(field_id)
        else:
            field_index = field_id

        # extract the field and the associated time series
        field_obj = self._field[field_index].copy()
        if label:
            field_obj.label = label
        field_slice = self._field._slices[field_index]
        data = [d[field_slice].reshape(field_obj.data.shape) for d in self.data]

        # create the corresponding MemoryStorage
        return MemoryStorage(
            times=self.times, data=data, field_obj=field_obj, info=self.info
        )

    def extract_time_range(
        self, t_range: Union[float, Tuple[float, float], None] = None
    ) -> "MemoryStorage":
        """extract a particular time interval

        Note:
            This might return a view into the original data, so modifying the returned
            data can also change the underlying original data.

        Args:
            t_range (float or tuple):
                Determines the range of time points included in the result. If only a
                single number is given, all data up to this time point are included.

        Returns:
            :class:`MemoryStorage`: a storage instance that contains the extracted data.
        """
        from .memory import MemoryStorage  # @Reimport

        # get the time bracket
        try:
            t_start, t_end = t_range  # type: ignore
        except TypeError:
            t_start, t_end = None, t_range
        if t_start is None:
            t_start = self.times[0]
        if t_end is None:
            t_end = self.times[-1]

        # determine the associated indices
        i_start = np.searchsorted(self.times, t_start, side="left")
        i_end = np.searchsorted(self.times, t_end, side="right")

        # extract the actual memory
        return MemoryStorage(
            times=self.times[i_start:i_end],
            data=self.data[i_start:i_end],
            field_obj=self._field,
            info=self.info,
        )

    def apply(
        self,
        func: Callable,
        out: Optional[StorageBase] = None,
        *,
        progress: bool = False,
    ) -> StorageBase:
        """applies function to each field in a storage

        Args:
            func (callable):
                The function to apply to each stored field. The function must either
                take as a single argument the field or as two arguments the field and
                the associated time point. In both cases, it should return a field.
            out (:class:`~pde.storage.base.StorageBase`):
                Storage to which the output is written. If omitted, a new
                :class:`~pde.storage.memory.MemoryStorage` is used and returned
            progress (bool):
                Flag indicating whether the progress is shown during the calculation

        Returns:
            :class:`~pde.storage.base.StorageBase`: The new storage that contains the
            data after the function `func` has been applied
        """
        # get the number of arguments that the user function expects
        num_args = len(signature(func).parameters)
        writing = False  # flag indicating whether output storage was opened

        for t, field in display_progress(
            self.items(), total=len(self), enabled=progress
        ):
            # apply the user function
            if num_args == 0:
                transformed = func()
            elif num_args == 1:
                transformed = func(field)
            else:
                transformed = func(field, t)

            if not isinstance(transformed, FieldBase):
                raise TypeError("The user function must return a field")

            if out is None:
                from .memory import MemoryStorage  # @Reimport

                out = MemoryStorage(field_obj=transformed)

            if not writing:
                out.start_writing(transformed)
                writing = True

            out.append(transformed, t)

        if writing:
            out.end_writing()  # type: ignore

        # make sure that a storage is returned, even when no fields are present
        if out is None:
            from .memory import MemoryStorage  # @Reimport

            out = MemoryStorage()

        return out

    def copy(
        self, out: Optional[StorageBase] = None, *, progress: bool = False
    ) -> StorageBase:
        """copies all fields in a storage to a new one

        Args:
            out (:class:`~pde.storage.base.StorageBase`):
                Storage to which the output is written. If omitted, a new
                :class:`~pde.storage.memory.MemoryStorage` is used and returned
            progress (bool):
                Flag indicating whether the progress is shown during the calculation

        Returns:
            :class:`~pde.storage.base.StorageBase`: The new storage that contains the
            copied data
        """
        # apply the identity function to do the copy
        return self.apply(lambda x: x, out=out, progress=progress)


class StorageTracker(TrackerBase):
    """Tracker that stores data in special storage classes

    Attributes:
        storage (:class:`~pde.storage.base.StorageBase`):
            The underlying storage class through which the data can be accessed
    """

    @fill_in_docstring
    def __init__(
        self,
        storage,
        interval: IntervalData = 1,
        *,
        transformation: Optional[Callable[[FieldBase, float], FieldBase]] = None,
    ):
        """
        Args:
            storage (:class:`~pde.storage.base.StorageBase`):
                Storage instance to which the data is written
            interval:
                {ARG_TRACKER_INTERVAL}
            transformation (callable, optional):
                A function that transforms the current state into a new field or field
                collection, which is then stored. This allows to store derived
                quantities of the field during calculations. The argument needs to be a
                callable function taking 1 or 2 arguments. The first argument always is
                the current field, while the optional second argument is the associated
                time.
        """
        super().__init__(interval=interval)
        self.storage = storage
        if transformation is not None and not callable(transformation):
            raise TypeError("`transformation` must be callable")
        self.transformation = transformation

    def _transform(self, field: FieldBase, t: float) -> FieldBase:
        """transforms the field according to the defined transformation"""
        if self.transformation is None:
            return field
        elif self.transformation.__code__.co_argcount == 1:
            return self.transformation(field)  # type: ignore
        else:
            return self.transformation(field, t)

    def initialize(self, field: FieldBase, info: Optional[InfoDict] = None) -> float:
        """
        Args:
            field (:class:`~pde.fields.FieldBase`):
                An example of the data that will be analyzed by the tracker
            info (dict):
                Extra information from the simulation

        Returns:
            float: The first time the tracker needs to handle data
        """
        result = super().initialize(field, info)
        self.storage.start_writing(self._transform(field, 0), info)
        return result

    def handle(self, field: FieldBase, t: float) -> None:
        """handle data supplied to this tracker

        Args:
            field (:class:`~pde.fields.FieldBase`):
                The current state of the simulation
            t (float): The associated time
        """
        self.storage.append(self._transform(field, t), time=t)

    def finalize(self, info: Optional[InfoDict] = None) -> None:
        """finalize the tracker, supplying additional information

        Args:
            info (dict):
                Extra information from the simulation
        """
        super().finalize(info)
        self.storage.end_writing()



"""
Spherically-symmetric grids in 2 and 3 dimensions. These are grids that only
discretize the radial direction, assuming symmetry with respect to all angles.
This choice implies that differential operators might not be applicable to all fields.
For instance, the divergence of a vector field on a spherical grid can only be
represented as a scalar field on the same grid if the θ-component of the vector field
vanishes.

.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
"""

from __future__ import annotations

from abc import ABCMeta
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Generator,
    Literal,
    Optional,
    Tuple,
    TypeVar,
    Union,
)

import numpy as np

from ..tools.cache import cached_property
from ..tools.plotting import plot_on_axes
from .base import (
    CoordsType,
    DimensionError,
    GridBase,
    _check_shape,
    discretize_interval,
)
from .cartesian import CartesianGrid

if TYPE_CHECKING:
    from .boundaries.axes import Boundaries  # @UnusedImport


π = np.pi
PI_4 = 4 * π
PI_43 = 4 / 3 * π


TNumArr = TypeVar("TNumArr", float, np.ndarray)


def volume_from_radius(radius: TNumArr, dim: int) -> TNumArr:
    """Return the volume of a sphere with a given radius

    Args:
        radius (float or :class:`~numpy.ndarray`):
            Radius of the sphere
        dim (int):
            Dimension of the space

    Returns:
        float or :class:`~numpy.ndarray`: Volume of the sphere
    """
    if dim == 1:
        return 2 * radius
    elif dim == 2:
        return π * radius**2
    elif dim == 3:
        return PI_43 * radius**3
    else:
        raise NotImplementedError(f"Cannot calculate the volume in {dim} dimensions")


class SphericalSymGridBase(GridBase, metaclass=ABCMeta):
    r"""Base class for d-dimensional spherical grids with angular symmetry

    The angular symmetry implies that states only depend on the radial
    coordinate :math:`r`, which is discretized uniformly as

    .. math::
        r_i = R_\mathrm{inner} + \left(i + \frac12\right) \Delta r
        \quad \text{for} \quad i = 0, \ldots, N - 1
        \quad \text{with} \quad
            \Delta r = \frac{R_\mathrm{outer} - R_\mathrm{inner}}{N}

    where :math:`R_\mathrm{outer}` is the outer radius of the grid and
    :math:`R_\mathrm{inner}` corresponds to a possible inner radius, which is
    zero by default. The radial direction is discretized by :math:`N` support
    points.
    """

    _periodic = [False]  # the radial axis is not periodic
    num_axes = 1  # the number of independent axes

    boundary_names = {"inner": (0, False), "outer": (0, True)}

    def __init__(
        self, radius: Union[float, Tuple[float, float]], shape: Union[Tuple[int], int]
    ):
        r"""
        Args:
            radius (float or tuple of floats):
                Radius :math:`R_\mathrm{outer}` in case a simple float is given. If a
                tuple is supplied it is interpreted as the inner and outer radius,
                :math:`(R_\mathrm{inner}, R_\mathrm{outer})`.
            shape (tuple or int):
                The number :math:`N` of support points along the radial coordinate.
        """
        super().__init__()
        shape_list = _check_shape(shape)
        if not len(shape_list) == 1:
            raise ValueError(f"`shape` must be a single number, not {shape_list}")
        self._shape: Tuple[int] = (int(shape_list[0]),)

        try:
            r_inner, r_outer = radius  # type: ignore
        except TypeError:
            r_inner, r_outer = 0, float(radius)  # type: ignore

        if r_inner < 0:
            raise ValueError("Inner radius must be positive")
        if r_inner >= r_outer:
            raise ValueError("Outer radius must be larger than inner radius")

        # radial discretization
        rs, dr = discretize_interval(r_inner, r_outer, self.shape[0])

        self._axes_coords = (rs,)
        self._axes_bounds = ((r_inner, r_outer),)
        self._discretization = np.array((dr,))

    @property
    def state(self) -> Dict[str, Any]:
        """state: the state of the grid"""
        return {"radius": self.radius, "shape": self.shape}

    @classmethod
    def from_state(cls, state: Dict[str, Any]) -> SphericalSymGridBase:  # type: ignore
        """create a field from a stored `state`.

        Args:
            state (dict):
                The state from which the grid is reconstructed.
        """
        state_copy = state.copy()
        obj = cls(radius=state_copy.pop("radius"), shape=state_copy.pop("shape"))
        if state_copy:
            raise ValueError(f"State items {state_copy.keys()} were not used")
        return obj

    @classmethod
    def from_bounds(  # type: ignore
        cls,
        bounds: Tuple[Tuple[float, float]],
        shape: Tuple[int],
        periodic: Tuple[bool],
    ) -> SphericalSymGridBase:
        """
        Args:
            bounds (tuple):
                Give the coordinate range for the radial axis.
            shape (tuple):
                The number of support points for the radial axis

        Returns:
            :class:`SphericalGridBase`: represents the region chosen by bounds
        """
        if len(bounds) != 1:
            raise ValueError(
                f"`bounds` must be given as ((r_min, r_max),). Got {bounds} instead"
            )
        return cls(bounds[0], shape)

    @property
    def has_hole(self) -> bool:
        """returns whether the inner radius is larger than zero"""
        return self.axes_bounds[0][0] > 0

    @property
    def radius(self) -> Union[float, Tuple[float, float]]:
        """float: radius of the sphere"""
        r_inner, r_outer = self.axes_bounds[0]
        if r_inner == 0:
            return r_outer
        else:
            return r_inner, r_outer

    @property
    def volume(self) -> float:
        """float: total volume of the grid"""
        r_inner, r_outer = self.axes_bounds[0]
        volume = volume_from_radius(r_outer, dim=self.dim)
        if r_inner > 0:
            volume -= volume_from_radius(r_inner, dim=self.dim)
        return volume

    @cached_property()
    def cell_volume_data(self) -> Tuple[np.ndarray]:
        """tuple of :class:`~numpy.ndarray`: the volumes of all cells"""
        dr = self.discretization[0]
        rs = self.axes_coords[0]
        volumes_h = volume_from_radius(rs + 0.5 * dr, dim=self.dim)
        volumes_l = volume_from_radius(rs - 0.5 * dr, dim=self.dim)
        return ((volumes_h - volumes_l).reshape(self.shape[0]),)

    def get_random_point(
        self,
        *,
        boundary_distance: float = 0,
        avoid_center: bool = False,
        coords: CoordsType = "cartesian",
        rng: Optional[np.random.Generator] = None,
    ) -> np.ndarray:
        """return a random point within the grid

        Note that these points will be uniformly distributed in the volume, implying
        they are not uniformly distributed on the radial axis.

        Args:
            boundary_distance (float):
                The minimal distance this point needs to have from all boundaries.
            avoid_center (bool):
                Determines whether the boundary distance should also be kept from the
                center, i.e., whether points close to the center are returned.
            coords (str):
                Determines the coordinate system in which the point is specified. Valid
                values are `cartesian`, `cell`, and `grid`;
                see :meth:`~pde.grids.base.GridBase.transform`.
            rng (:class:`~numpy.random.Generator`):
                Random number generator (default: :func:`~numpy.random.default_rng()`)

        Returns:
            :class:`~numpy.ndarray`: The coordinates of the point
        """
        if rng is None:
            rng = np.random.default_rng()

        # handle the boundary distance
        r_inner, r_outer = self.axes_bounds[0]
        r_min = r_inner + boundary_distance if avoid_center else r_inner
        r_max = r_outer - boundary_distance
        if r_max <= r_min:
            raise RuntimeError("Random points would be too close to boundary")

        # choose random radius scaled such that points are uniformly distributed
        r = np.array(
            [rng.uniform(r_min**self.dim, r_max**self.dim) ** (1 / self.dim)]
        )
        if coords == "cartesian":
            # choose random angles for the already chosen radius
            if self.dim == 2:
                φ = rng.uniform(0, 2 * np.pi)
                point = np.r_[r, φ]
            elif self.dim == 3:
                θ = np.arccos(rng.uniform(-1, 1))
                φ = rng.uniform(0, 2 * np.pi)
                point = np.r_[r, θ, φ]
            else:
                raise NotImplementedError(f"{self.dim} dimensions")

            return self.point_to_cartesian(point, full=True)

        elif coords == "cell":
            return self.transform(r, "grid", "cell")

        elif coords == "grid":
            return r

        else:
            raise ValueError(f"Unknown coordinate system `{coords}`")

    def get_line_data(self, data: np.ndarray, extract: str = "auto") -> Dict[str, Any]:
        """return a line cut along the radial axis

        Args:
            data (:class:`~numpy.ndarray`):
                The values at the grid points
            extract (str):
                Determines which cut is done through the grid. This parameter is mainly
                supplied for a consistent interface and has no effect for polar grids.

        Returns:
            A dictionary with information about the line cut, which is
            convenient for plotting.
        """
        if extract not in {"auto", "r", "radial"}:
            raise ValueError(f"Unknown extraction method `{extract}`")

        return {
            "data_x": self.axes_coords[0],
            "data_y": data,
            "extent_x": self.axes_bounds[0],
            "label_x": self.axes[0],
        }

    def get_image_data(
        self,
        data: np.ndarray,
        performance_goal: Literal["speed", "quality"] = "speed",
        fill_value: float = 0,
        masked: bool = True,
    ) -> Dict[str, Any]:
        """return a 2d-image of the data

        Args:
            data (:class:`~numpy.ndarray`):
                The values at the grid points
            performance_goal (str):
                Determines the method chosen for interpolation. Possible options are
                `speed` and `quality`.
            fill_value (float):
                The value assigned to invalid positions (those inside the hole or
                outside the region).
            masked (bool):
                Whether a :class:`numpy.ma.MaskedArray` is returned for the data instead
                of the normal :class:`~numpy.ndarray`.

        Returns:
            :dict: A dictionary with information about the image, which is  convenient
            for plotting.
        """
        from scipy import interpolate

        _, r_outer = self.axes_bounds[0]
        r_data = self.axes_coords[0]

        if self.has_hole:
            num = int(np.ceil(r_outer / self.discretization[0]))
            x_positive, _ = discretize_interval(0, r_outer, num)
        else:
            x_positive = r_data

        x = np.r_[-x_positive[::-1], x_positive]
        xs, ys = np.meshgrid(x, x, indexing="ij")
        r_img = np.hypot(xs, ys)

        if performance_goal == "speed":
            # interpolate over the new coordinates using linear interpolation
            f = interpolate.interp1d(
                r_data,
                data,
                copy=False,
                bounds_error=False,
                fill_value=fill_value,
                assume_sorted=True,
            )
            data_int = f(r_img.flat).reshape(r_img.shape)

        elif performance_goal == "quality":
            # interpolate over the new coordinates using radial base function
            f = interpolate.Rbf(r_data, data, function="cubic")
            data_int = f(r_img)

        else:
            raise ValueError(f"Performance goal `{performance_goal}` undefined")

        if masked:
            mask = (r_img < r_data[0]) | (r_data[-1] < r_img)
            data_int = np.ma.masked_array(data_int, mask=mask)

        return {
            "data": data_int,
            "x": x,
            "y": x,
            "extent": (-r_outer, r_outer, -r_outer, r_outer),
            "label_x": "x",
            "label_y": "y",
        }

    def iter_mirror_points(
        self, point: np.ndarray, with_self: bool = False, only_periodic: bool = True
    ) -> Generator:
        """generates all mirror points corresponding to `point`

        Args:
            point (:class:`~numpy.ndarray`):
                The point within the grid
            with_self (bool):
                Whether to include the point itself
            only_periodic (bool):
                Whether to only mirror along periodic axes

        Returns:
            A generator yielding the coordinates that correspond to mirrors
        """
        if with_self:
            yield np.asanyarray(point, dtype=np.double)

    def point_from_cartesian(self, points: np.ndarray) -> np.ndarray:
        """convert points given in Cartesian coordinates to this grid

        Args:
            points (:class:`~numpy.ndarray`):
                Points given in Cartesian coordinates.

        Returns:
            :class:`~numpy.ndarray`: Points given in the coordinates of the grid
        """
        points = np.atleast_1d(points)
        assert points.shape[-1] == self.dim, f"Point must have {self.dim} coordinates"
        return np.linalg.norm(points, axis=-1, keepdims=True)  # type: ignore

    def polar_coordinates_real(
        self, origin=None, *, ret_angle: bool = False, **kwargs
    ) -> Union[np.ndarray, Tuple[np.ndarray, ...]]:
        """return spherical coordinates associated with the grid

        Args:
            origin:
                Place holder variable to comply with the interface
            ret_angle (bool):
                Determines whether angles are returned alongside the distance. If
                `False` only the distance to the origin is returned for each support
                point of the grid. If `True`, the distance and angles are returned. Note
                that in the case of spherical grids, this angle is zero by convention.
        """
        # check the consistency of the origin argument, which can be set for other grids
        if origin is not None:
            origin = np.array(origin, dtype=np.double, ndmin=1)
            if not np.array_equal(origin, np.zeros(self.dim)):
                raise RuntimeError(f"Origin must be {str([0]*self.dim)}")

        # the distance to the origin is exactly the radial coordinate
        rs = self.axes_coords[0]
        if ret_angle:
            return (rs,) + (np.zeros_like(rs),) * (self.dim - 1)
        else:
            return rs

    def get_cartesian_grid(
        self,
        mode: Literal["valid", "inscribed", "full", "circumscribed"] = "valid",
        num: Optional[int] = None,
    ) -> CartesianGrid:
        """return a Cartesian grid for this spherical one

        Args:
            mode (str):
                Determines how the grid is determined. Setting it to 'valid' (or
                'inscribed') only returns points that are fully resolved in the
                spherical grid, e.g., the Cartesian grid is inscribed in the sphere.
                Conversely, 'full' (or 'circumscribed') returns all data, so the
                Cartesian grid is circumscribed.
            num (int):
                Number of support points along each axis of the returned grid.

        Returns:
            :class:`pde.grids.cartesian.CartesianGrid`: The requested grid
        """
        # pick how the grid is determined
        if mode == "valid":
            if self.has_hole:
                self._logger.warning("Sphere has holes, so not all points are valid")
            bounds = self.radius / np.sqrt(self.dim)
        elif mode == "inscribed":
            bounds = self.radius / np.sqrt(self.dim)
        elif mode == "full" or mode == "circumscribed":
            bounds = self.radius
        else:
            raise ValueError(f"Unsupported mode `{mode}`")

        # determine the grid points
        if num is None:
            num = 2 * round(bounds / self.discretization[0])
        grid_bounds = [(-bounds, bounds)] * self.dim
        return CartesianGrid(grid_bounds, num)

    @plot_on_axes()
    def plot(self, ax, **kwargs):
        r"""visualize the spherically symmetric grid in two dimensions

        Args:
            {PLOT_ARGS}
            \**kwargs:
                Extra arguments are passed on the to the matplotlib plotting routines,
                e.g., to set the color of the lines
        """
        from matplotlib import collections, patches

        kwargs.setdefault("edgecolor", kwargs.get("color", "k"))
        kwargs.setdefault("facecolor", "none")
        (rb,) = self.axes_bounds
        rmax = rb[1]

        # draw circular parts
        circles = []
        for r in np.linspace(*rb, self.shape[0] + 1):
            if r == 0:
                c = patches.Circle((0, 0), 0.01 * rmax)
            else:
                c = patches.Circle((0, 0), r)
            circles.append(c)
        ax.add_collection(collections.PatchCollection(circles, **kwargs))

        ax.set_xlim(-rmax, rmax)
        ax.set_xlabel("x")
        ax.set_ylim(-rmax, rmax)
        ax.set_ylabel("y")
        ax.set_aspect(1)


class PolarSymGrid(SphericalSymGridBase):
    r"""2-dimensional polar grid assuming angular symmetry

    The angular symmetry implies that states only depend on the radial
    coordinate :math:`r`, which is discretized uniformly as

    .. math::
        r_i = R_\mathrm{inner} + \left(i + \frac12\right) \Delta r
        \quad \text{for} \quad i = 0, \ldots, N - 1
        \quad \text{with} \quad
            \Delta r = \frac{R_\mathrm{outer} - R_\mathrm{inner}}{N}

    where :math:`R_\mathrm{outer}` is the outer radius of the grid and
    :math:`R_\mathrm{inner}` corresponds to a possible inner radius, which is
    zero by default. The radial direction is discretized by :math:`N` support
    points.
    """

    dim = 2  # dimension of the described space
    axes = ["r"]
    axes_symmetric = ["phi"]
    coordinate_constraints = [0, 1]  # axes not described explicitly

    def point_to_cartesian(
        self, points: np.ndarray, *, full: bool = False
    ) -> np.ndarray:
        """convert coordinates of a point to Cartesian coordinates

        This function returns points along the y-coordinate, i.e, the x
        coordinates will be zero.

        Args:
            points (:class:`~numpy.ndarray`):
                The grid coordinates of the points
            full (bool):
                Flag indicating whether angular coordinates are specified

        Returns:
            :class:`~numpy.ndarray`: The Cartesian coordinates of the point
        """
        points = np.atleast_1d(points)

        if full:
            # angles are supplied
            if points.shape[-1] != self.dim:
                raise DimensionError(f"Shape {points.shape} cannot denote points")

            x = points[..., 0] * np.cos(points[..., 1])
            y = points[..., 0] * np.sin(points[..., 1])
        else:
            # angles are not supplied
            if points.shape[-1] != self.num_axes:
                raise DimensionError(f"Shape {points.shape} cannot denote points")

            y = points[..., 0]
            x = np.zeros_like(y)

        return np.stack((x, y), axis=-1)


class SphericalSymGrid(SphericalSymGridBase):
    r"""3-dimensional spherical grid assuming spherical symmetry

    The symmetry implies that states only depend on the radial coordinate
    :math:`r`, which is discretized as follows:

    .. math::
        r_i = R_\mathrm{inner} + \left(i + \frac12\right) \Delta r
        \quad \text{for} \quad i = 0, \ldots, N - 1
        \quad \text{with} \quad
            \Delta r = \frac{R_\mathrm{outer} - R_\mathrm{inner}}{N}

    where :math:`R_\mathrm{outer}` is the outer radius of the grid and
    :math:`R_\mathrm{inner}` corresponds to a possible inner radius, which is
    zero by default. The radial direction is discretized by :math:`N` support
    points.

    Warning:
        Not all results of differential operators on vectorial and tensorial fields can
        be expressed in terms of fields that only depend on the radial coordinate
        :math:`r`. In particular, the gradient of a vector field can only be calculated
        if the azimuthal component of the vector field vanishes. Similarly, the
        divergence of a tensor field can only be taken in special situations.
    """

    dim = 3  # dimension of the described space
    axes = ["r"]
    axes_symmetric = ["theta", "phi"]
    coordinate_constraints = [0, 1, 2]  # axes not described explicitly

    def point_to_cartesian(
        self, points: np.ndarray, *, full: bool = False
    ) -> np.ndarray:
        """convert coordinates of a point to Cartesian coordinates

        This function returns points along the z-coordinate, i.e, the x and y
        coordinates will be zero.

        Args:
            points (:class:`~numpy.ndarray`):
                The grid coordinates of the points
            full (bool):
                Flag indicating whether angular coordinates are specified

        Returns:
            :class:`~numpy.ndarray`: The Cartesian coordinates of the point
        """
        points = np.atleast_1d(points)
        if full:
            # angles are supplied
            if points.shape[-1] != self.dim:
                raise DimensionError(f"Shape {points.shape} cannot denote points")

            rsinθ = points[..., 0] * np.sin(points[..., 1])
            x = rsinθ * np.cos(points[..., 2])
            y = rsinθ * np.sin(points[..., 2])
            z = points[..., 0] * np.cos(points[..., 1])

        else:
            # angles are not supplied
            if points.shape[-1] != self.num_axes:
                raise DimensionError(f"Shape {points.shape} cannot denote points")
            z = points[..., 0]
            x = y = np.zeros_like(z)

        return np.stack((x, y, z), axis=-1)



"""
Defines an explicit solver using multiprocessing via MPI
   
.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de> 
"""

from typing import Callable, List, Literal, Union

import numpy as np
from numba.extending import register_jitable

from ..fields.base import FieldBase
from ..grids._mesh import GridMesh
from ..pdes.base import PDEBase
from ..tools import mpi
from ..tools.math import OnlineStatistics
from ..tools.typing import BackendType
from .explicit import ExplicitSolver


class ExplicitMPISolver(ExplicitSolver):
    """class for solving partial differential equations explicitly using MPI

    Warning:
        This solver can only be used if MPI is properly installed. In particular, python
        scripts then need to be started using :code:`mpirun` or :code:`mpiexec`. Please
        refer to the documentation of your MPI distribution for details.

    The main idea of the solver is to take the full initial state in the main node
    (ID 0) and split the grid into roughly equal subgrids. The main node then
    distributes these subfields to all other nodes and each node creates the right hand
    side of the PDE for itself (and independently). Each node then advances the PDE
    independently, ensuring proper coupling to neighboring nodes via special boundary
    conditions, which exchange field values between sub grids. This is implemented by
    the :meth:`get_boundary_conditions` method of the sub grids, which takes the
    boundary conditions for the full grid and creates conditions suitable for the
    specific sub grid on the given node. The trackers (and thus all input and output)
    are only handled on the main node.

    Warning:
        The function providing the right hand side of the PDE needs to support MPI. This
        is automatically the case for local evaluations (which only use the field value
        at the current position), for the differential operators provided by :mod:`pde`,
        and integration of fields. Similarly, `modify_after_step` can only be used to do
        local modifications since the field data supplied to the function is local to
        each MPI node.

    Example:
        A minimal example using the MPI solver is

        .. code-block:: python

           from pde import DiffusionPDE, ScalarField, UnitGrid

           grid = UnitGrid([64, 64])
           state = ScalarField.random_uniform(grid, 0.2, 0.3)

           eq = DiffusionPDE(diffusivity=0.1)
           result = eq.solve(state, t_range=10, dt=0.1, method="explicit_mpi")

           if result is not None:  # restrict the output to the main node
              result.plot()

        Saving this script as `multiprocessing.py`, a parallel simulation is started by

        .. code-block:: bash

            mpiexec -n 2 python3 multiprocessing.py

        Here, the number `2` determines the number of cores that will be used. Note that
        macOS might require an additional hint on how to connect the processes even
        when they are run on the same machine (e.g., your workstation). It might help to
        run :code:`mpiexec -n 2 -host localhost python3 multiprocessing.py` in this case
    """

    name = "explicit_mpi"

    def __init__(
        self,
        pde: PDEBase,
        scheme: Literal["euler", "runge-kutta", "rk", "rk45"] = "euler",
        decomposition: Union[int, List[int]] = -1,
        *,
        backend: BackendType = "auto",
        adaptive: bool = False,
        tolerance: float = 1e-4,
    ):
        """
        Args:
            pde (:class:`~pde.pdes.base.PDEBase`):
                The instance describing the pde that needs to be solved
            scheme (str):
                Defines the explicit scheme to use. Supported values are 'euler' and
                'runge-kutta' (or 'rk' for short).
            decomposition (list of ints):
                Number of subdivision in each direction. Should be a list of length
                `grid.num_axes` specifying the number of nodes along this axis. If one
                value is `-1`, its value will be determined from the number of available
                nodes. The default value decomposed the first axis using all available
                nodes.
            backend (str):
                Determines how the function is created. Accepted  values are 'numpy` and
                'numba'. Alternatively, 'auto' lets the code decide for the most optimal
                backend.
            adaptive (bool):
                When enabled, the time step is adjusted during the simulation using the
                error tolerance set with `tolerance`.
            tolerance (float):
                The error tolerance used in adaptive time stepping. This is used in
                adaptive time stepping to choose a time step which is small enough so
                the truncation error of a single step is below `tolerance`.
        """
        super().__init__(
            pde, scheme=scheme, backend=backend, adaptive=adaptive, tolerance=tolerance
        )
        self.decomposition = decomposition

    def _make_error_synchronizer(self) -> Callable[[float], float]:
        """return helper function that synchronizes errors between multiple processes"""
        if mpi.parallel_run:
            # in a parallel run, we need to return the maximal error
            from ..tools.mpi import mpi_allreduce

            @register_jitable
            def synchronize_errors(error: float) -> float:
                """return maximal error accross all cores"""
                return mpi_allreduce(error, "MAX")  # type: ignore

            return synchronize_errors  # type: ignore
        else:
            return super()._make_error_synchronizer()

    def make_stepper(
        self, state: FieldBase, dt=None
    ) -> Callable[[FieldBase, float, float], float]:
        """return a stepper function using an explicit scheme

        Args:
            state (:class:`~pde.fields.base.FieldBase`):
                An example for the state from which the grid and other information can
                be extracted
            dt (float):
                Time step of the explicit stepping. If `None`, this solver specifies
                1e-3 as a default value.

        Returns:
            Function that can be called to advance the `state` from time `t_start` to
            time `t_end`. The function call signature is `(state: numpy.ndarray,
            t_start: float, t_end: float)`
        """
        if not mpi.parallel_run:
            self._logger.warning(
                "Using `ExplicitMPISolver` without a proper multiprocessing run. "
                "Scripts need to be started with `mpiexec` to profit from multiple cores"
            )

        if dt is None:
            # support `None` as a default value, so the controller can signal that
            # the solver should use a default time step
            dt = 1e-3
            if not self.adaptive:
                self._logger.warning(
                    "Explicit stepper with a fixed time step did not receive any "
                    f"initial value for `dt`. Using dt={dt}, but specifying a value or "
                    "enabling adaptive stepping is advisable."
                )

        self.info["dt"] = dt
        self.info["dt_adaptive"] = self.adaptive
        self.info["steps"] = 0
        self.info["stochastic"] = self.pde.is_sde
        self.info["state_modifications"] = 0.0
        self.info["use_mpi"] = True
        self.info["scheme"] = self.scheme

        # decompose the state into multiple cells
        self.mesh = GridMesh.from_grid(state.grid, self.decomposition)
        sub_state = self.mesh.extract_subfield(state)
        self.info["grid_decomposition"] = self.mesh.shape

        if self.adaptive:
            # create stepper with adaptive steps
            self.info["dt_statistics"] = OnlineStatistics()
            adaptive_stepper = self._make_adaptive_stepper(sub_state)

            def wrapped_stepper(
                state: FieldBase, t_start: float, t_end: float
            ) -> float:
                """advance `state` from `t_start` to `t_end` using adaptive steps"""
                nonlocal dt  # `dt` stores value for the next call

                # distribute the end time and the field to all nodes
                t_end = self.mesh.broadcast(t_end)
                substate_data = self.mesh.split_field_data_mpi(state.data)

                # evolve the sub state
                t_last, dt, steps, modifications = adaptive_stepper(
                    substate_data, t_start, t_end, dt, self.info["dt_statistics"]
                )

                # check whether dt is the same for all processes
                dt_list = self.mesh.allgather(dt)
                if not np.isclose(min(dt_list), max(dt_list)):
                    # abort simulations in all nodes when they went out of sync
                    raise RuntimeError(f"Processes went out of sync: dt={dt_list}")

                # collect the data from all nodes
                modification_list = self.mesh.gather(modifications)
                self.mesh.combine_field_data_mpi(substate_data, out=state.data)

                if mpi.is_main:
                    self.info["steps"] += steps
                    self.info["state_modifications"] += sum(modification_list)  # type: ignore
                return t_last

        else:
            # create stepper with fixed steps
            fixed_stepper = self._make_fixed_stepper(sub_state, dt)

            def wrapped_stepper(
                state: FieldBase, t_start: float, t_end: float
            ) -> float:
                """advance `state` from `t_start` to `t_end` using fixed steps"""
                # calculate number of steps (which is at least 1)
                steps = max(1, int(np.ceil((t_end - t_start) / dt)))

                # distribute the number of steps and the field to all nodes
                steps = self.mesh.broadcast(steps)
                substate_data = self.mesh.split_field_data_mpi(state.data)

                # evolve the sub state
                t_last, modifications = fixed_stepper(substate_data, t_start, steps)

                # check whether t_last is the same for all processes
                t_list = self.mesh.gather(t_last)
                if t_list is not None and not np.isclose(min(t_list), max(t_list)):
                    raise RuntimeError(f"Processes went out of sync: t_last={t_list}")

                # collect the data from all nodes
                modification_list = self.mesh.gather(modifications)
                self.mesh.combine_field_data_mpi(substate_data, out=state.data)

                # store information in the main node
                if mpi.is_main:
                    self.info["steps"] += steps
                    self.info["state_modifications"] += sum(modification_list)  # type: ignore
                return t_last

        return wrapped_stepper
